from .action_types import (RUN_VALIDATION_REPORT, SET_OPENBADGES_VERSION, SET_VALIDATION_SUBJECT,
                           SET_VERIFIED_PROFILE)


def run_validation_report():
    return {
        'type': RUN_VALIDATION_REPORT
    }


def set_openbadges_version(version_string):
    """
    Set the version of the node that is the subject of the present validation request/report.
    :param version_string: One of ("0.5", "1.0", "1.1", "2.0")
    :return: dict
    """
    known_versions = ("0.5", "1.0", "1.1", "2.0")

    if version_string not in known_versions:
        raise ValueError("version {} not among known versions")

    return {
        'type': SET_OPENBADGES_VERSION,
        'version': version_string
    }


def set_validation_subject(node_id):
    """
    Set the ID of the node in the graph that is the subject of the present validation request/report.
    :param node_id: 
    :return: dict
    """
    return {
        'type': SET_VALIDATION_SUBJECT,
        'node_id': node_id
    }


def set_verified_recipient_profile(id_type, confirmed_id):
    """
    :param type: str
    :param value: str
    :return: dict
    """
    return {
        'type': SET_VERIFIED_PROFILE,
        'identityType': id_type,
        'identityValue': confirmed_id
    }



import base64
import json
from openbadges_bakery import unbake
from pyld import jsonld
import re
from tempfile import NamedTemporaryFile

from ..actions.input import set_input_type, store_input
from ..actions.tasks import add_task, report_message
from ..actions.validation_report import set_validation_subject
from ..exceptions import TaskPrerequisitesError
from ..openbadges_context import OPENBADGES_CONTEXT_V2_URI
from ..tasks.utils import is_url
from ..utils import jsonld_use_cache, make_string_from_bytes, MESSAGE_LEVEL_ERROR
from ..tasks.task_types import DETECT_INPUT_TYPE, FETCH_HTTP_NODE, PROCESS_JWS_INPUT
from .utils import abbreviate_value as abv, task_result


"""
Helpful utils
"""

def input_is_json(user_input):
    try:
        value = json.loads(make_string_from_bytes(user_input))
        return True
    except (ValueError, TypeError):
        return False


def input_is_jws(user_input):
    jws_regex = re.compile(r'^[A-z0-9\-=]+.[A-z0-9\-=]+.[A-z0-9\-_=]+$')
    return bool(jws_regex.match(make_string_from_bytes(user_input)))


def find_id_in_jsonld(json_string, jsonld_options):
    input_data = json.loads(json_string)
    result = jsonld.compact(input_data, OPENBADGES_CONTEXT_V2_URI, options=jsonld_options)
    node_id = result.get('id','')
    return node_id


def find_1_0_verify_url(json_string, options):
    input_data = json.loads(json_string)
    try:
        return input_data['verify']['url']
    except KeyError:
        return ''

"""
Input-processing tasks
"""
def detect_input_type(state, task_meta, **options):
    """
    Detects what data format user has provided and saves to the store.
    """
    input_value = state.get('input').get('value')
    depth = task_meta.get('depth')
    detected_type = None
    new_actions = []

    if is_url(input_value):
        detected_type = 'url'
        new_actions.append(set_input_type(detected_type))
        new_actions.append(add_task(
            FETCH_HTTP_NODE, url=input_value, is_potential_baked_input=task_meta.get('is_potential_baked_input', True), depth=depth
        ))
        new_actions.append(set_validation_subject(input_value))
    elif input_is_json(input_value):
        for url_finder in [find_id_in_jsonld, find_1_0_verify_url]:
            id_url = url_finder(input_value, options.get('jsonld_options', jsonld_use_cache))
            if is_url(id_url):
                detected_type = 'url'
                new_actions.append(store_input(id_url))
                break
            else:
                detected_type = 'json'

        new_actions.append(set_input_type(detected_type))
        if detected_type == 'json':
            new_actions.append(report_message(
                'Could not determine verifiable input from provided JSON. No hosted verification URL found.',
                message_level=MESSAGE_LEVEL_ERROR
            ))
        elif detected_type == 'url':
            new_actions.append(add_task(FETCH_HTTP_NODE, url=id_url, is_potential_baked_input=False, depth=depth))
            new_actions.append(set_validation_subject(id_url))
    elif input_is_jws(input_value):
        detected_type = 'jws'
        new_actions.append(set_input_type(detected_type))
        new_actions.append(add_task(PROCESS_JWS_INPUT, data=input_value, depth=depth))
    else:
        raise NotImplementedError("only URL, JSON, or JWS input implemented so far")

    return task_result(
        message="Input of type {} detected.".format(detected_type),
        actions=new_actions
    )


def process_baked_resource(state, task_meta, **options):
    try:
        node_id = task_meta['node_id']
        resource_b64 = state['input']['original_json'][node_id]
        depth = task_meta['depth']
    except KeyError:
        raise TaskPrerequisitesError()
    try:
        search_string = resource_b64.decode('utf-8')
        match = re.search(r'^data:(image\/png|image\/svg\+xml);base64,(.+)$', search_string)
        content_type = match.group(1)
        suffix = '.png' if content_type == 'image/png' else '.svg'
        b64_data = match.group(2)
    except (AttributeError, IndexError):
        return task_result(False, "Cannot determine image type or content from dataURI {}".format(abv(resource_b64)))

    baked_file = NamedTemporaryFile(suffix=suffix)
    baked_file.write(base64.b64decode(b64_data))
    baked_file.seek(0)

    assertion_data = unbake(baked_file)

    if assertion_data:
        actions = [
            store_input(assertion_data),
            add_task(DETECT_INPUT_TYPE, is_potential_baked_input=False, depth=depth)
        ]
        return task_result(True, "Retrieved baked data from image resource {}".format(abv(node_id)), actions)
    else:
        return task_result(
            False, "Resource {} was an image of known type but no baked Open Badges data was available".format(
                abv(node_id)
            ))



import unittest

from openbadges.verifier.actions.tasks import add_task
from openbadges.verifier.tasks import task_named
from openbadges.verifier.tasks.task_types import VALIDATE_EXPECTED_NODE_CLASS
from openbadges.verifier.tasks.validation import OBClasses


class ValidateLanguagePropertyTests(unittest.TestCase):
    def test_validate_language_prop_basic(self):
        options = {'max_validation_depth': 3}

        badgeclass = {
            'id': 'http://example.org/badgeclass',
            '@language': 'en-US'
        }
        state = {'graph': [badgeclass]}
        task = add_task(VALIDATE_EXPECTED_NODE_CLASS, node_id=badgeclass['id'],
                        expected_class=OBClasses.BadgeClass, depth=0)
        result, message, actions = task_named(VALIDATE_EXPECTED_NODE_CLASS)(state, task, **options)
        self.assertTrue(result)

        l_actions = [a for a in actions if a.get('prop_name') == '@language']
        self.assertEqual(len(l_actions), 1)

        result, message, actions = task_named(l_actions[0]['name'])(state, l_actions[0], **options)
        self.assertTrue(result)



import os
import responses

from openbadges.verifier.openbadges_context import OPENBADGES_CONTEXT_V2_URI

from .testfiles.test_components import test_components


# Make sure to decorate calling function with @responses.activate
def set_up_context_mock():
    context_data = test_components['openbadges_context']
    responses.add(
        responses.GET,
        OPENBADGES_CONTEXT_V2_URI,
        body=context_data,
        status=200,
        content_type='application/ld+json')


def set_up_image_mock(url):
    with open(os.path.join(os.path.dirname(__file__), 'testfiles', 'public_domain_heart.png'), 'rb') as f:
        responses.add(responses.GET, url, body=f.read(), content_type='image/png')



from .crypto import (process_jws_input, verify_jws_signature, verify_key_ownership,
                     verify_signed_assertion_not_revoked,)
from .extensions import validate_extension_node, validate_single_extension
from .images import validate_image
from .input import detect_input_type, process_baked_resource
from .graph import fetch_http_node, flatten_refetch_embedded_resource, intake_json, jsonld_compact_data
from .object_upgrades import upgrade_0_5_node, upgrade_1_0_node, upgrade_1_1_node
from .validation import (assertion_timestamp_checks, assertion_verification_dependencies,
                         criteria_property_dependencies, detect_and_validate_node_class,
                         identity_object_property_dependencies, issuer_property_dependencies, placeholder_task,
                         validate_expected_node_class, validate_rdf_type_property, validate_property,
                         validate_revocationlist_entries, )
from .verification import (hosted_id_in_verification_scope, verify_recipient_against_trusted_profile)
from .task_types import *


FUNCTIONS = {
    ASSERTION_TIMESTAMP_CHECKS:                assertion_timestamp_checks,
    ASSERTION_VERIFICATION_CHECK:              placeholder_task,
    ASSERTION_VERIFICATION_DEPENDENCIES:       assertion_verification_dependencies,
    DETECT_AND_VALIDATE_NODE_CLASS:            detect_and_validate_node_class,
    DETECT_INPUT_TYPE:                         detect_input_type,
    CRITERIA_PROPERTY_DEPENDENCIES:            criteria_property_dependencies,
    FETCH_HTTP_NODE:                           fetch_http_node,
    FLATTEN_EMBEDDED_RESOURCE:                 flatten_refetch_embedded_resource,
    HOSTED_ID_IN_VERIFICATION_SCOPE:           hosted_id_in_verification_scope,
    JSONLD_COMPACT_DATA:                       jsonld_compact_data,
    IDENTITY_OBJECT_PROPERTY_DEPENDENCIES:     identity_object_property_dependencies,
    INTAKE_JSON:                               intake_json,
    ISSUER_PROPERTY_DEPENDENCIES:              issuer_property_dependencies,
    PROCESS_BAKED_RESOURCE:                    process_baked_resource,
    PROCESS_JWS_INPUT:                         process_jws_input,
    UPGRADE_0_5_NODE:                          upgrade_0_5_node,
    UPGRADE_1_0_NODE:                          upgrade_1_0_node,
    UPGRADE_1_1_NODE:                          upgrade_1_1_node,
    VALIDATE_EXPECTED_NODE_CLASS:              validate_expected_node_class,
    VALIDATE_EXTENSION_NODE:                   validate_extension_node,
    VALIDATE_EXTENSION_SINGLE:                 validate_single_extension,
    IMAGE_VALIDATION:                          validate_image,
    VALIDATE_RDF_TYPE_PROPERTY:                validate_rdf_type_property,
    VALIDATE_PROPERTY:                         validate_property,
    VALIDATE_REVOCATIONLIST_ENTRIES:           validate_revocationlist_entries,
    VERIFY_JWS:                                verify_jws_signature,
    VERIFY_KEY_OWNERSHIP:                      verify_key_ownership,
    VERIFY_RECIPIENT_IDENTIFIER:               verify_recipient_against_trusted_profile,
    VERIFY_SIGNED_ASSERTION_NOT_REVOKED:       verify_signed_assertion_not_revoked,
}


def task_named(key):
    return FUNCTIONS[key]


def run_task(state, task):
    return task_named(task['name'])(state, task)



from pydux import combine_reducers

from .input import input_reducer
from .graph import graph_reducer
from .tasks import task_reducer
from .verification_report import verification_report_reducer


main_reducer = combine_reducers({
    'input': input_reducer,
    'graph': graph_reducer,
    'tasks': task_reducer,
    'report': verification_report_reducer
})



import os.path
from requests_cache import CachedSession
import responses
import unittest

from openbadges.verifier.actions.tasks import add_task
from openbadges.verifier.actions.action_types import STORE_ORIGINAL_RESOURCE
from openbadges.verifier.reducers.input import input_reducer
from openbadges.verifier.tasks import task_named
from openbadges.verifier.tasks.validation import OBClasses
from openbadges.verifier.tasks.task_types import (IMAGE_VALIDATION, VALIDATE_EXPECTED_NODE_CLASS)
from openbadges.verifier.tasks.utils import is_data_uri
from openbadges.verifier.utils import CachableDocumentLoader


class ImageValidationTests(unittest.TestCase):
    @responses.activate
    def test_validate_badgeclass_image_formats(self):
        session = CachedSession(backend='memory', expire_after=100000)
        loader = CachableDocumentLoader(use_cache=True, session=session)
        options = {
            'jsonld_options': {'documentLoader': loader},
            'max_validation_depth': 3
        }
        image_url = 'http://example.org/awesomebadge.png'
        badgeclass = {
            'id': 'http://example.org/badgeclass',
            'name': 'Awesome badge',
            'image': image_url
        }
        state = {'graph': [badgeclass]}

        with open(os.path.join(os.path.dirname(__file__), 'testfiles', 'public_domain_heart.png'), 'rb') as f:
            responses.add(responses.GET, badgeclass['image'], body=f.read(), content_type='image/png')
        response = session.get(badgeclass['image'])
        self.assertEqual(response.status_code, 200)

        task_meta = add_task(
            VALIDATE_EXPECTED_NODE_CLASS, node_id=badgeclass['id'], expected_class=OBClasses.BadgeClass, depth=0)

        result, message, actions = task_named(VALIDATE_EXPECTED_NODE_CLASS)(state, task_meta, **options)
        self.assertTrue(result)

        image_task = [a for a in actions if a.get('prop_name') == 'image'][0]
        class_image_validation_task = [a for a in actions if a.get('name') == IMAGE_VALIDATION][0]
        result, message, actions = task_named(image_task['name'])(state, image_task, **options)
        self.assertTrue(result)
        self.assertEqual(len(actions), 0)

        result, message, actions = task_named(class_image_validation_task['name'])(
            state, class_image_validation_task, **options)
        self.assertTrue(result)
        self.assertEqual(len(actions), 1)
        # self.assertEqual(actions[0]['name'], STORE_ORIGINAL_RESOURCE)

        # Case 2: Embedded image document
        badgeclass['image'] = {
            'id': 'http://example.org/awesomebadge.png',
            'author': 'http://someoneelse.org/1',
            'caption': 'A hexagon with attitude'
        }

        # Validate BadgeClass, queuing the image node validation task
        result, message, actions = task_named(image_task['name'])(state, image_task, **options)
        self.assertTrue(result)
        self.assertEqual(len(actions), 1, "Image node validation task queued")

        # Run image node task discovery
        next_task = actions[0]
        result, message, actions = task_named(next_task['name'])(state, next_task, **options)
        self.assertTrue(result)

        # Run validation task for the Image node
        next_task = [a for a in actions if a.get('name') == IMAGE_VALIDATION][0]
        result, message, actions = task_named(next_task['name'])(state, next_task, **options)
        self.assertTrue(result)

        # Store image data
        next_task = actions[0]
        self.assertEqual(next_task['type'], STORE_ORIGINAL_RESOURCE)
        new_state = input_reducer({}, next_task)
        self.assertTrue(new_state['original_json'][image_url].startswith('data:'), "Data is stored in the expected spot.")

    def test_base64_data_uri_in_badgeclass(self):
        data_uri = 'data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mOUMyqsBwACeQFChxl' \
                   'ltgAAAABJRU5ErkJggg=='
        self.assertTrue(is_data_uri(data_uri))

        session = CachedSession(backend='memory', expire_after=100000)
        loader = CachableDocumentLoader(use_cache=True, session=session)
        options = {
            'jsonld_options': {'documentLoader': loader},
            'max_validation_depth': 3
        }
        badgeclass = {
            'id': 'http://example.org/badgeclass',
            'name': 'Awesome badge',
            'image': data_uri
        }
        state = {'graph': [badgeclass]}

        task_meta = add_task(
            VALIDATE_EXPECTED_NODE_CLASS, node_id=badgeclass['id'], expected_class=OBClasses.BadgeClass, depth=0)

        result, message, actions = task_named(VALIDATE_EXPECTED_NODE_CLASS)(state, task_meta, **options)
        self.assertTrue(result)

        image_task = [a for a in actions if a.get('prop_name') == 'image'][0]
        class_image_validation_task = [a for a in actions if a.get('name') == IMAGE_VALIDATION][0]
        result, message, actions = task_named(image_task['name'])(state, image_task, **options)
        self.assertTrue(result)
        self.assertEqual(len(actions), 0)

        result, message, actions = task_named(class_image_validation_task['name'])(
            state, class_image_validation_task, **options)
        self.assertTrue(result)
        self.assertEqual(len(actions), 0, "There is no need to separately store the data_uri")

        # Case 2 SVG
        badgeclass['image'] = 'data:image/svg+xml;charset=utf-8;base64,PHN2ZyBoZWlnaHQ9IjEwMCIgd2lkdGg9IjEwMCI+DQogID' \
                              'xjaXJjbGUgY3g9IjUwIiBjeT0iNTAiIHI9IjQwIiBzdHJva2U9ImJsYWNrIiBzdHJva2Utd2lkdGg9IjMiIGZp' \
                              'bGw9IiM1NTU1ZmYiIC8+ICANCjwvc3ZnPg=='

        result, message, actions = task_named(class_image_validation_task['name'])(
            state, class_image_validation_task, **options)
        self.assertTrue(result)
        self.assertEqual(len(actions), 0, "There is no need to separately store the data_uri")

        # Case 3 naughty JPG
        badgeclass['image'] = 'data:image/jpeg;base64,/9j/4AAQSkZJRgABAQEAYABgAAD/4QAWRXhpZgAASUkqAAgAAAAAAAAAAAD/2wB' \
                              'DAAEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBA' \
                              'QH/2wBDAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQE' \
                              'BAQEBAQH/wAARCAABAAEDASIAAhEBAxEB/8QAFQABAQAAAAAAAAAAAAAAAAAAAAr/xAAUEAEAAAAAAAAAAAAAA' \
                              'AAAAAAA/8QAFAEBAAAAAAAAAAAAAAAAAAAAAP/EABQRAQAAAAAAAAAAAAAAAAAAAAD/2gAMAwEAAhEDEQA/AL+' \
                              'AAf/Z'

        result, message, actions = task_named(class_image_validation_task['name'])(
            state, class_image_validation_task, **options)
        self.assertFalse(result)

    def test_base64_data_uri_in_assertion(self):
        data_uri = 'data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mOUMyqsBwACeQFChxl' \
                   'ltgAAAABJRU5ErkJggg=='
        self.assertTrue(is_data_uri(data_uri))

        session = CachedSession(backend='memory', expire_after=100000)
        loader = CachableDocumentLoader(use_cache=True, session=session)
        options = {
            'jsonld_options': {'documentLoader': loader},
            'max_validation_depth': 3
        }
        assertion = {
            'id': 'http://example.org/assertion',
            'image': data_uri
        }
        state = {'graph': [assertion]}

        task_meta = add_task(
            VALIDATE_EXPECTED_NODE_CLASS, node_id=assertion['id'], expected_class=OBClasses.Assertion, depth=0)

        result, message, actions = task_named(VALIDATE_EXPECTED_NODE_CLASS)(state, task_meta, **options)
        self.assertTrue(result)

        image_task = [a for a in actions if a.get('prop_name') == 'image'][0]
        class_image_validation_task = [a for a in actions if a.get('name') == IMAGE_VALIDATION][0]
        result, message, actions = task_named(image_task['name'])(state, image_task, **options)
        self.assertFalse(result, "The Assertion image property validation task should not allow data URIs")
        self.assertEqual(len(actions), 0)

        result, message, actions = task_named(class_image_validation_task['name'])(
            state, class_image_validation_task, **options)
        self.assertFalse(result, "The image validation task should also fail for the same reason")



import unittest

from pydux import create_store

from openbadges.verifier.reducers import main_reducer
from openbadges.verifier.actions.tasks import add_task, resolve_task
from openbadges.verifier.reducers.tasks import _new_state_with_updated_item
from openbadges.verifier.tasks.utils import abbreviate_value
from openbadges.verifier.state import INITIAL_STATE, filter_active_tasks


class TaskActionTests(unittest.TestCase):
    def setUp(self):
        self.store = create_store(main_reducer, INITIAL_STATE)

    def test_add_task_basic(self):
        self.store.dispatch(add_task('DETECT_INPUT_TYPE', **{'other': 123}))
        tasks = self.store.get_state().get('tasks')
        self.assertEqual(len(tasks), 1)
        self.assertEqual(tasks[0]['task_id'], 1)
        self.assertEqual(tasks[0]['name'], 'DETECT_INPUT_TYPE')
        self.assertEqual(tasks[0]['other'], 123)

        self.store.dispatch(add_task('DETECT_INPUT_TYPE'))
        tasks = self.store.get_state().get('tasks')
        self.assertEqual(len(tasks), 2)
        self.assertEqual(tasks[1]['task_id'], 2)

    def test_cannot_add_unknown_task(self):
        with self.assertRaises(AssertionError):
            self.store.dispatch(add_task('UNKNOWN_SOLDIER'))

    def test_resolve_task(self):
        self.store.dispatch(add_task('DETECT_INPUT_TYPE', **{'other': 123}))
        tasks = self.store.get_state().get('tasks')
        self.assertEqual(len(filter_active_tasks(self.store.get_state())), 1)

        self.store.dispatch(resolve_task(tasks[0]['task_id']))
        tasks = self.store.get_state().get('tasks')
        self.assertTrue(tasks[0]['complete'])
        self.assertEqual(len(filter_active_tasks(self.store.get_state())), 0)


    def test_task_updater_internals(self):
        initial = [{'task_id': 1}, {'task_id': 2}, {'task_id': 3}]
        updated = _new_state_with_updated_item(initial, 2, {'task_id': 2, 'foo': 'bar'})
        self.assertEqual(updated[1].get('foo'), 'bar')


class TaskUtilsTests(unittest.TestCase):
    def test_abbreviate_values(self):
        def chars(length):
            return ''.join(['a' for _ in range(length)])

        self.assertEqual(abbreviate_value('abc'), 'abc')
        self.assertEqual(abbreviate_value(chars(47)), chars(47))
        self.assertEqual(abbreviate_value(chars(48)), chars(48) + '...')

        test_list = ['interesting', 'entries']
        self.assertEqual(abbreviate_value(test_list), 'interesting, entries')
        self.assertEqual(abbreviate_value(['interesting']), 'interesting')
        self.assertEqual(abbreviate_value(['interesting', 1]), 'interesting, 1')
        self.assertEqual(abbreviate_value([{'a': 'b'}, 'interesting']), "{'a': 'b'}, interesting")



import os
from openbadges_bakery import bake
import responses
import unittest
from pydux import create_store

from openbadges.verifier.actions.validation_report import set_validation_subject
from openbadges.verifier.reducers import main_reducer
from openbadges.verifier.state import INITIAL_STATE
from openbadges.verifier.verifier import generate_report, verification_store

try:
    from .testfiles.test_components import test_components
    from tests.utils import set_up_context_mock
except (ImportError, SystemError):
    from .testfiles.test_components import test_components
    from .testutils import set_up_context_mock


class VerificationReportTests(unittest.TestCase):
    @staticmethod
    def set_response_mocks():
        # Make sure to add @responses.activate decorator in calling method
        set_up_context_mock()
        responses.add(
            responses.GET, 'https://example.org/beths-robotics-badge.json',
            body=test_components['2_0_basic_assertion'], status=200,
            content_type='application/ld+json'
        )
        responses.add(
            responses.GET, 'https://example.org/robotics-badge.json',
            body=test_components['2_0_basic_badgeclass'], status=200,
            content_type='application/ld+json'
        )
        responses.add(
            responses.GET, 'https://example.org/organization.json',
            body=test_components['2_0_basic_issuer'], status=200,
            content_type='application/ld+json'
        )

    def test_validation_subject_reducer(self):
        """
        The validationReport should contain an ID of the root node that was validated in the graph.
        For example, if the input was the URL of an Assertion, that URL (the assertion['id'] would appear.
        """
        node_id = 'http://example.com/assertion'
        store = create_store(main_reducer, INITIAL_STATE)
        store.dispatch(set_validation_subject(node_id))
        state = store.get_state()
        self.assertEqual(state['report']['validationSubject'], node_id)

    @responses.activate
    def test_subject_set_from_badge_input(self):
        url = 'https://example.org/beths-robotics-badge.json'
        self.set_response_mocks()
        store = verification_store(url)
        report = generate_report(store)
        self.assertEqual(report['report']['validationSubject'], url)

    @responses.activate
    def test_subject_set_from_badge_image(self):
        with open(os.path.join(os.path.dirname(__file__), 'testfiles', 'public_domain_heart.png'), 'rb') as f:
            image = bake(f, test_components['2_0_basic_assertion'])
            self.set_response_mocks()
            store = verification_store(image)
            report = generate_report(store)
            self.assertEqual(report['report']['validationSubject'], 'https://example.org/beths-robotics-badge.json')

    @responses.activate
    def test_confirmed_recipient_profile_reported(self):
        url = 'https://example.org/beths-robotics-badge.json'
        email = 'nobody@example.org'
        self.set_response_mocks()
        store = verification_store(url, recipient_profile={'email': email})
        report = generate_report(store)
        self.assertEqual(report['report']['recipientProfile']['email'], email)

    def test_validation_version(self):
        """
        The validationReport reports which version of the Open Badges spec the validationSubject was found to be in.
        prop: "openBadgesversion"
        """
        pass


    def test_messages_warnings_counts(self):
        """
        The validationReport contains the properties of messages, warningCount, errorCount, valid
        :return: 
        """
        pass



from ..actions.action_types import (ADD_TASK, DELETE_OUTDATED_NODE_TASKS, REPORT_MESSAGE, RESOLVE_TASK,
                                    TRIGGER_CONDITION, UPDATE_TASK)
from ..tasks.task_types import (FETCH_HTTP_NODE, VALIDATE_EXPECTED_NODE_CLASS, VALIDATE_PROPERTY,
                                VALIDATE_RDF_TYPE_PROPERTY, UPGRADE_TASKS)
from ..state import filter_active_tasks, MESSAGE_LEVEL_INFO


def _task_to_add_exists(state, action):
    try:
        if action.get('name') == VALIDATE_EXPECTED_NODE_CLASS:
            task = [t for t in state if
                action['name'] == t.get('name') and
                action.get('node_id') == t.get('node_id') and
                action.get('node_path') == t.get('node_path')
            ][0]

        elif action.get('name') in [VALIDATE_PROPERTY, VALIDATE_RDF_TYPE_PROPERTY]:
            task = [t for t in state if
                    action.get('node_id') == t.get('node_id') and
                    action.get('node_path') == t.get('node_path') and
                    action.get('prop_name') == t.get('prop_name')][0]

        elif action.get('name') == FETCH_HTTP_NODE:
            task = [
                t for t in state if
                action.get('url') == t.get('url')
            ][0]

        else:
            return False
    except IndexError:
        return False

    return True


def task_reducer(state=None, action=None):
    if state is None or len(state) == 0:
        state = []
        task_counter = 1
    else:
        task_counter = state[-1]['task_id'] + 1

    if action.get('type') == ADD_TASK and not _task_to_add_exists(state, action):
        new_task = {'task_id': task_counter, 'complete': False}
        for key in [k for k in list(action.keys()) if k != 'type']:
            new_task[key] = action[key]
        return list(state) + [new_task]

    elif action.get('type') == RESOLVE_TASK:
        try:
            task = [t for t in state if t['task_id'] == action['task_id']][0]
        except KeyError:
            return state
        else:
            update = task.copy()
            update.update({
                'complete': True,
                'success': action.get('success'),
                'result': action.get('result')
            })
            return _new_state_with_updated_item(state, action['task_id'], update)

    elif action.get('type') == UPDATE_TASK:
        try:
            task = [t for t in state if t['task_id'] == action['task_id']][0]
        except KeyError:
            return state
        else:
            update = task.copy()
            for key in [k for k in list(action.keys()) if k not in ('type', 'task_id',)]:
                update[key] = action[key]
            return _new_state_with_updated_item(state, action['task_id'], update)

    elif action.get('type') == REPORT_MESSAGE:
        new_task = {
            'task_id': task_counter,
            'name': REPORT_MESSAGE,
            'complete': True,
            'success': action.get('success', True),
            'result': action.get('message'),
            'messageLevel': action.get('messageLevel', MESSAGE_LEVEL_INFO)
        }
        return list(state) + [new_task]

    elif action.get('type') == TRIGGER_CONDITION:
        new_task = {
            'task_id': task_counter,
            'complete': True,
            'name': action.get('name'),
            'success': action.get('success', True),
            'result': action.get('message')
        }
        return list(state) + [new_task]

    elif action.get('type') == DELETE_OUTDATED_NODE_TASKS:
        node_id = action.get('node_id', 'UNKNOWN')
        return list(
            filter(lambda tt: not all([tt.get('node_id') == node_id and tt.get('name') in UPGRADE_TASKS]), state)
        )
    return state


def _new_state_with_updated_item(state, item_id, update):
    new_state = []
    for i in range(0, len(state)):
        if item_id != state[i].get('task_id'):
            new_state.append(state[i])
        else:
            new_state.append(update)
    return new_state



import rfc3986
import six

from ..actions.tasks import report_message
from ..actions.validation_report import set_verified_recipient_profile
from ..exceptions import TaskPrerequisitesError
from ..state import get_node_by_id, get_node_by_path
from ..utils import identity_hash, list_of, MESSAGE_LEVEL_WARNING

from .utils import abbreviate_value as abv, task_result


def _default_allowed_origins_for_issuer_id(issuer_id):
    return rfc3986.uri_reference(issuer_id).authority


def _default_verification_policy(issuer_node):
    issuer_id = issuer_node.get('id')
    return {
        'type': 'VerificationObject',
        'allowedOrigins': _default_allowed_origins_for_issuer_id(issuer_id),
        'verificationProperty': 'id'
    }


def hosted_id_in_verification_scope(state, task_meta, **options):
    try:
        assertion_id = task_meta.get('node_id')
        assertion_node = get_node_by_id(state, assertion_id)

        badgeclass_node = get_node_by_id(state, assertion_node['badge'])
        issuer_node = get_node_by_id(state, badgeclass_node['issuer'])
        issuer_id = issuer_node.get('id')
    except IndexError:
        raise TaskPrerequisitesError()

    try:
        verification_policy = issuer_node.get('verification')
        if verification_policy is None or isinstance(verification_policy, six.string_types):
            verification_policy = get_node_by_id(state, issuer_node.get('verification'))
    except IndexError:
        verification_policy = _default_verification_policy(issuer_node)

    if verification_policy.get('startsWith'):
        starts_with = list_of(verification_policy['startsWith'])
        if not any([assertion_id.startswith(i) for i in starts_with]):
            return task_result(
                False, "Assertion id {}".format(assertion_id) +
                "does not start with any permitted values in its issuer's verification policy."
            )

    allowed_origins = list_of(
        verification_policy.get(
            'allowedOrigins', _default_allowed_origins_for_issuer_id(issuer_id))
    )
    if not allowed_origins or not issuer_id.startswith('http'):
        return task_result(
            True, "There are no allowed hosted verification URL domains for this issuer's verification policy.",
            actions=[report_message(
                'Issuer {} has no HTTP domain to enforce hosted verification policy against.'.format(issuer_id),
                message_level=MESSAGE_LEVEL_WARNING)]
        )
    elif rfc3986.uri_reference(assertion_id).authority not in allowed_origins:
        return task_result(
            False, 'Assertion {} not hosted in allowed origins {}'.format(
                abv(assertion_id), abv(allowed_origins))
        )

    return task_result(
        True, 'Assertion {} origin matches allowed value in issuer verification policy {}.'.format(
            abv(assertion_id), abv(allowed_origins))
    )


def _matches_hash(profile_identifier, id_hash, salt=''):
    if id_hash.startswith('md5'):
        return identity_hash(profile_identifier, salt, alg='md5') == id_hash
    elif id_hash.startswith('sha256'):
        return identity_hash(profile_identifier, salt, alg='sha256') == id_hash

    raise TypeError("Cannot interpret hash type of {}".format(id_hash))


def verify_recipient_against_trusted_profile(state, task_meta, **options):
    try:
        # Use the ID of the first Assertion found in current state
        assertion_id = [n for n in state['graph'] if n.get('type') == 'Assertion'][0]['id']

        identity_node = get_node_by_path(state, [assertion_id, 'recipient'])
        profile_id = task_meta['node_id']
        profile_node = get_node_by_id(state, profile_id)
    except (IndexError, KeyError):
        raise TaskPrerequisitesError()

    actions = []

    a = identity_node['identity']  # Recipient value in (a)ssertion
    recipient_type = identity_node['type']
    if recipient_type not in ['id', 'email', 'url', 'telephone']:
        actions += [report_message(
            "Recipient identifier type {} in assertion {} is not one of the recommended types".format(
                recipient_type, assertion_id))]

    try:
        p = list_of(profile_node[recipient_type])  # matching type recipient value in submitted (p)rofile
    except KeyError:
        return task_result(
            False, "Profile identifier property of type {} not found in submitted profile {}".format(
                recipient_type, profile_id
            ), actions)

    if identity_node['hashed']:
        salt = identity_node.get('salt', '')
        for possible_id in p:
            if _matches_hash(possible_id, a.lower(), salt):
                confirmed_id = possible_id
                break
        else:
            # If no identifier in the profile matches, return failure
            return task_result(
                False,
                "Profile {} identifier(s) {} of type {} did not match assertion {} recipient hash {}.".format(
                    abv(profile_id), abv(p), recipient_type, abv(assertion_id), a),
                actions)
    elif a in p:
        confirmed_id = a
    else:
        return task_result(
            False,
            "Profile {} identifier {} of type {} did not match assertion {} recipient value {}".format(
                abv(profile_id), p, recipient_type, abv(assertion_id), a),
            actions)

    actions.append(set_verified_recipient_profile(recipient_type, confirmed_id))
    return task_result(True, "Assertion {} awarded to trusted profile identifier {} of type {}".format(
        abv(assertion_id), confirmed_id, recipient_type), actions)



import json
from openbadges_bakery import unbake
from pydux import create_store
import traceback

from .actions.input import set_input_type, store_input
from .actions.tasks import add_task, report_message, resolve_task, trigger_condition
from .exceptions import SkipTask, TaskPrerequisitesError
from .logger import logger
from .openbadges_context import OPENBADGES_CONTEXT_V2_URI
from .reducers import main_reducer
from .state import (filter_active_tasks, filter_messages_for_report, format_message,
                    INITIAL_STATE, MESSAGE_LEVEL_ERROR, MESSAGE_LEVEL_WARNING,)
from . import tasks
from .tasks.task_types import INTAKE_JSON, JSONLD_COMPACT_DATA, VALIDATE_EXTENSION_NODE
from .tasks.validation import OBClasses
from .utils import list_of, CachableDocumentLoader, jsonld_use_cache



DEFAULT_OPTIONS = {
    'include_original_json': False,  # Return the original JSON strings fetched from HTTP
    'use_cache': True,
    'cache_backend': 'memory',
    'cache_expire_after': 300,
    'jsonld_options': jsonld_use_cache,
    'max_validation_depth': 3
}


def _get_options(options):
    if options:
        selected = DEFAULT_OPTIONS.copy()
        selected.update(options)
    else:
        selected = DEFAULT_OPTIONS

    if selected['use_cache']:
        doc_loader = CachableDocumentLoader(
            use_cache=selected['use_cache'],
            backend=selected['cache_backend'],
            expire_after=selected['cache_expire_after']
        )
    else:
        doc_loader = CachableDocumentLoader(use_cache=False)

    selected['jsonld_options'] = {'documentLoader': doc_loader}
    return selected


def call_task(task_func, task_meta, store, options=DEFAULT_OPTIONS):
    """
    Calls and resolves a task function in response to a queued task. May result
    in additional actions added to the queue.
    :param task_func: func
    :param task_meta: dict (single entry in tasks state)
    :param store: pydux store
    :return:
    """
    actions = []
    try:
        success, message, actions = task_func(store.get_state(), task_meta, **options)
    except SkipTask:
        raise NotImplemented("Implement SkipTask handling in call_task")
    except TaskPrerequisitesError:
        message = "Task could not run due to unmet prerequisites."
        store.dispatch(resolve_task(task_meta.get('task_id'), success=False, result=message))
    except Exception as e:
        error_message = traceback.format_exception_only(type(e), e)
        logger.error(traceback.format_exc())
        message = "{} {}".format(e.__class__, error_message)
        store.dispatch(resolve_task(task_meta.get('task_id'), success=False, result=message))
    else:
        store.dispatch(resolve_task(task_meta.get('task_id'), success=success, result=message))
        if success:
            for trigger in list_of(task_meta.get('triggers_completion', [])):
                store.dispatch(trigger_condition(trigger, 'Completed by {}: {}'.format(
                    task_meta.get('task_id'), task_meta.get('name')
                )))

    # Make updates and queue up next tasks.
    for action in actions:
        if not isinstance(action, dict):
            raise TypeError("Task {} returned actions of an unreadable type. Task details: {}".format(
                task_meta.get('name'), json.dumps(task_meta)
            ))
        store.dispatch(action)


def verification_store(badge_input, recipient_profile=None, store=None, options=DEFAULT_OPTIONS):
    if store is None:
        store = create_store(main_reducer, INITIAL_STATE)
    try:
        if hasattr(badge_input, 'read') and hasattr(badge_input, 'seek'):
            badge_input.seek(0)
            badge_data = unbake(badge_input)
            if not badge_data:
                raise ValueError("Could not find Open Badges metadata in file.")
        else:
            badge_data = badge_input
    except ValueError as e:
        # Could not obtain badge data from input. Set the result as a failed DETECT_INPUT_TYPE task.
        store.dispatch(store_input(badge_input.name))
        store.dispatch(add_task(tasks.DETECT_INPUT_TYPE))
        store.dispatch(set_input_type('file'))
        task = store.get_state()['tasks'][0]
        store.dispatch(resolve_task(task.get('task_id'), success=False, result=e.message))
    else:
        store.dispatch(store_input(badge_data))
        store.dispatch(add_task(tasks.DETECT_INPUT_TYPE, depth=0))

    if recipient_profile:
        profile_id = recipient_profile.get('id')
        recipient_profile['@context'] = recipient_profile.get('@context', OPENBADGES_CONTEXT_V2_URI)
        task = add_task(
            JSONLD_COMPACT_DATA,
            data=json.dumps(recipient_profile),
            expected_class=OBClasses.ExpectedRecipientProfile,
            depth=0
        )
        if profile_id:
            task['node_id'] = profile_id
        store.dispatch(task)

    last_task_id = 0
    while len(filter_active_tasks(store.get_state())):
        active_tasks = filter_active_tasks(store.get_state())
        task_meta = active_tasks[0]
        task_func = tasks.task_named(task_meta['name'])

        if task_meta['task_id'] == last_task_id:
            break

        last_task_id = task_meta['task_id']
        call_task(task_func, task_meta, store, options)

    return store


def generate_report(store, options=DEFAULT_OPTIONS):
    """
    Returns a report of validity information based on a store and its tasks.
    """
    state = store.get_state()

    processed_input = state['input'].copy()
    if not options.get('include_original_json'):
        try:
            del processed_input['original_json']
        except KeyError:
            pass

    tasks_for_messages_list = filter_messages_for_report(state)
    report = state['report'].copy()
    report['messages'] = []
    for task in tasks_for_messages_list:
        report['messages'].append(format_message(task))

    report['errorCount'] = len([m for m in report['messages'] if m['messageLevel'] == MESSAGE_LEVEL_ERROR])
    report['warningCount'] = len([m for m in report['messages'] if m['messageLevel'] == MESSAGE_LEVEL_WARNING])

    is_valid = True  # Assume to be true at first
    if bool(report['errorCount']) or len(state.get('graph', [])) == 0:
        is_valid = False
    report['valid'] = is_valid

    ret = {
        'graph': state['graph'],
        'input': processed_input,
        'report': report
    }
    return ret


def verify(badge_input, recipient_profile=None, **options):
    """
    Verify and validate Open Badges
    :param badge_input: str (url or json) or python file-like object (baked badge image)
    :param recipient_profile: dict of a trusted Profile describing the entity assumed to be recipient
    :param options: dict of options. See DEFAULT_OPTIONS for values
    :return: dict
    """
    selected_options = _get_options(options)
    store = verification_store(badge_input, recipient_profile, options=selected_options)
    return generate_report(store, options=selected_options)


def extension_validation_store(extension_input, store=None, options=DEFAULT_OPTIONS):
    if store is None:
        store = create_store(main_reducer, INITIAL_STATE)

    if not isinstance(extension_input, dict):
        raise ValueError

    store.dispatch(store_input(extension_input.copy()))

    extension_input['@context'] = extension_input.get('@context', OPENBADGES_CONTEXT_V2_URI)
    extension_input['id'] = extension_input.get('id', '_:extension_validation_input')
    compact_task = add_task(JSONLD_COMPACT_DATA, detectAndValidateClass=False, data=json.dumps(extension_input))
    store.dispatch(compact_task)

    tasks_remaining = True
    while tasks_remaining:
        active_tasks = filter_active_tasks(store.get_state())
        if len(active_tasks) < 1:
            tasks_remaining = False
            break
        task_meta = active_tasks[0]
        task_func = tasks.task_named(task_meta['name'])
        call_task(task_func, task_meta, store, options)


    all_tasks = store.get_state()['tasks']
    try:
        first_extension_node_validation_task = [t for t in all_tasks if t['name'] == VALIDATE_EXTENSION_NODE][0]
    except IndexError:
        store.dispatch(report_message(
            "No extensions were found to test. Check for proper use of context and type to declare an extension.",
            message_level=MESSAGE_LEVEL_ERROR, success=False
        ))

    return store


def validate_extensions(extension_input, **options):
    """
    Validate Open Badges Extensions

    :param extension_input: object with openbadges extension properites
    :param options: dict of options. See DEFAULT_OPTIONS for values
    :return: dict
    """
    selected_options = _get_options(options)
    store = extension_validation_store(extension_input, options=selected_options)
    return generate_report(store, selected_options)



from .action_types import ADD_NODE, PATCH_NODE, PATCH_NODE_REFERENCE, UPDATE_NODE
import sys


def add_node(node_id=None, data=None):
    if data is None:
        data = {}

    return {
        'type': ADD_NODE,
        'node_id': node_id,
        'data': data
    }


def update_node(node_id, data):
    action = {
        'type': UPDATE_NODE,
        'node_id': node_id,
        'data': data
    }
    return action


def patch_node(node_id, data):
    action = update_node(node_id, data)
    action['type'] = PATCH_NODE
    return action


def patch_node_reference(node_path, new_id):
    """
    When a reference to another node in the graph needs to be updated
    after discovering that it has a different canonical id, use this.
    :param node_path: list
    :param new_id: str
    :return: dict
    """
    return {
        'type': PATCH_NODE_REFERENCE,
        'node_path': node_path,
        'new_id': new_id
    }



import json
import responses
import unittest

from openbadges.verifier.actions.action_types import ADD_NODE, STORE_ORIGINAL_RESOURCE
from openbadges.verifier.actions.graph import add_node, patch_node, patch_node_reference
from openbadges.verifier.actions.tasks import add_task
from openbadges.verifier.reducers.graph import graph_reducer
from openbadges.verifier.state import get_node_by_id
from openbadges.verifier.tasks.graph import fetch_http_node, jsonld_compact_data
from openbadges.verifier.tasks import run_task
from openbadges.verifier.tasks.task_types import (DETECT_AND_VALIDATE_NODE_CLASS, FETCH_HTTP_NODE, INTAKE_JSON,
                                         JSONLD_COMPACT_DATA)
from openbadges.verifier.openbadges_context import OPENBADGES_CONTEXT_V2_URI
from openbadges.verifier.utils import MESSAGE_LEVEL_WARNING
from openbadges.verifier.verifier import verify

from .utils import set_up_context_mock, set_up_image_mock


try:
    from .testfiles.test_components import test_components
except (ImportError, SystemError):
    from .testfiles.test_components import test_components


class HttpFetchingTests(unittest.TestCase):

    @responses.activate
    def test_basic_http_fetch_task(self):
        url = 'http://example.com/assertionmaybe'
        responses.add(
            responses.GET, url,
            body=test_components['2_0_basic_assertion'],
            status=200, content_type='application/ld+json'
        )
        task = add_task(FETCH_HTTP_NODE, url=url, depth=0)

        success, message, actions = fetch_http_node({}, task)

        self.assertTrue(success)
        self.assertEqual(len(actions), 2)
        self.assertEqual(actions[0]['type'], STORE_ORIGINAL_RESOURCE)
        self.assertEqual(actions[1]['name'], INTAKE_JSON)

    @responses.activate
    def test_svg_fetch_with_complex_mimetype(self):
        url = 'http://example.com/circle'
        svg_circle = u'<svg height="100" width="100"><circle cx="50" cy="50" r="40" stroke="black" stroke-width="3" ' \
                      u'fill="red" /></svg>'
        responses.add(
            responses.GET, url,
            body=svg_circle,
            status=200, content_type='image/svg+xml; charset=utf-8'
        )
        task = add_task(FETCH_HTTP_NODE, url=url, depth=0)

        success, message, actions = fetch_http_node({}, task)

        self.assertTrue(success)
        self.assertEqual(len(actions), 1)
        self.assertEqual(actions[0]['type'], STORE_ORIGINAL_RESOURCE)


class NodeStorageTests(unittest.TestCase):
    def test_single_node_store(self):
        new_node = {
            "key1": 1
        }
        state = graph_reducer([], add_node('http://example.com/node1', new_node))
        self.assertEqual(len(state), 1)
        self.assertEqual(state[0]['id'], 'http://example.com/node1')
        self.assertEqual(state[0]['key1'], 1)

    def new_node_successfully_appends_to_state(self):
        new_node = {
            "key1": 1
        }
        state = graph_reducer([{"id": "_:b9000"}], add_node('http://example.com/node1', new_node))
        self.assertEqual(len(state), 2)

    def test_store_nested(self):
        new_node = {
            "key1": 1,
            "nested1": {"key2": 2}
        }
        state = graph_reducer([], add_node('http://example.com/node1', new_node))
        self.assertEqual(len(state), 1)
        first_node = [node for node in state if node['id'] == 'http://example.com/node1'][0]
        self.assertEqual(first_node['key1'], 1)
        nested_node = first_node['nested1']
        self.assertEqual(nested_node['key2'], 2)
        self.assertIsNone(nested_node.get('id'))

    def test_store_node_inaccurate_id_value(self):
        """
        Due to redirects, we may not have the canonical id for a node.
        If there's a conflict due to the id and the node[id] in add_node(id, node), 
        what should we do?
        """
        pass

    def test_store_with_included_blank_node_ids_in_input(self):
        """
        Make sure the graph reducer can handle assigning node ids when we include
        some collisions in the input that use the blank node identifier scheme _:
        """
        pass

    def test_store_lists(self):
        new_node = {
            "key1": 1,
            "b_list": [
                "b",
                {"c": 3}
            ]
        }

        state = graph_reducer([], add_node('http://example.com/node1', new_node))
        self.assertEqual(len(state), 1)
        root_node = get_node_by_id({'graph': state}, 'http://example.com/node1')
        self.assertEqual(root_node['id'], 'http://example.com/node1')
        self.assertEqual(root_node['key1'], 1)

        nested_node = root_node['b_list'][1]
        self.assertEqual(nested_node['c'], 3)


class NodeUpdateTests(unittest.TestCase):
    def test_patch_node(self):
        first_node = {'id': '_:b0', 'name': 'One'}
        second_node = {'id': '_:b1', 'name': 'Two'}

        action = patch_node(first_node['id'], {'name': 'New One'})
        state = graph_reducer([first_node, second_node], action)

        self.assertEqual(len(state), 2)
        updated_node = get_node_by_id({'graph': state}, first_node['id'])
        self.assertEqual(updated_node['name'], 'New One')

    def test_update_node_id(self):
        """
        TODO: allow ability to update a node id, replacing all references to that
        id in other nodes with the new id
        """
        pass

    def test_update_node_reference(self):
        first_node = {'id': '_:b0', 'badge': 'http://example.org/old'}
        graph_state = [first_node]
        new_id = 'http://example.org/new'
        action = patch_node_reference([first_node['id'], 'badge'], new_id)

        graph_state = graph_reducer(graph_state, action)
        self.assertEqual(graph_state[0]['badge'], new_id)



class JsonLdCompactTests(unittest.TestCase):
    def setUpContextCache(self):
        data = test_components['openbadges_context']
        responses.add(
            responses.GET, OPENBADGES_CONTEXT_V2_URI,
            body=data, status=200, content_type='application/ld+json'
        )

    @responses.activate
    def test_compact_node(self):
        self.setUpContextCache()

        data = """{
            "@context": {"thing_we_call_you_by": "http://schema.org/name"},
            "thing_we_call_you_by": "Test Data"
        }"""

        task = add_task(JSONLD_COMPACT_DATA, data=data, node_id='http://example.com/1')

        result, message, actions = jsonld_compact_data({}, task)
        self.assertTrue(result, "JSON-LD Compaction should be successful.")
        self.assertEqual(message, "Successfully compacted node http://example.com/1")
        self.assertEqual(
            len(actions), 2,
            "Should queue up add_node and add_task for type detection")
        self.assertEqual(
            actions[0]['data']['name'], "Test Data",
            "Node should be compacted into OB Context and use OB property names.")

    @responses.activate
    def test_no_task_data(self):
        task = add_task(JSONLD_COMPACT_DATA)

        result, message, actions = jsonld_compact_data({}, task)
        self.assertFalse(result)
        self.assertEqual(message, 'Could not load data')


    @responses.activate
    def test_reduce_compacted_output(self):
        self.setUpContextCache()

        data = {
            "@context": {"thing_we_call_you_by": "http://schema.org/name"},
            "thing_we_call_you_by": "Test Data"
        }

        task = add_task(JSONLD_COMPACT_DATA, data=json.dumps(data), node_id='_:b100')

        result, message, actions = jsonld_compact_data({}, task)

        state = graph_reducer([], actions[0])
        self.assertEqual(len(state), 1, "Node should be added to graph")
        self.assertEqual(state[0]['name'], data['thing_we_call_you_by'])
        self.assertEqual(state[0].get('id'), '_:b100', "Node should have a blank id assigned")


class ObjectRedirectionTests(unittest.TestCase):
    @responses.activate
    def test_node_has_id_different_from_fetch_url(self):
        set_up_context_mock()

        first_url = 'http://example.org/url1'
        second_url = 'http://example.org/url2'
        node_data = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': first_url,
            'name': 'Some Badge'
        }

        responses.add(responses.GET, first_url, json=node_data)
        responses.add(responses.GET, second_url, json=node_data)

        task = add_task(FETCH_HTTP_NODE, url=second_url, depth=0)
        state = {'graph': []}

        result, message, actions = run_task(state, task)
        self.assertTrue(result)
        self.assertEqual(len(actions), 2)
        self.assertEqual(actions[1]['name'], INTAKE_JSON)

        result, message, actions = run_task(state, actions[1])  # INTAKE_JSON
        self.assertTrue(result)
        self.assertEqual(len(actions), 2)
        self.assertEqual(actions[1]['name'], JSONLD_COMPACT_DATA)

        result, message, actions = run_task(state, actions[1])  # JSONLD_COMPACT_DATA
        self.assertTrue(result)
        self.assertEqual(len(actions), 3)
        self.assertEqual(actions[2]['name'], FETCH_HTTP_NODE)
        self.assertEqual(actions[2]['url'], first_url)
        self.assertEqual(actions[0]['messageLevel'], MESSAGE_LEVEL_WARNING)

        # Pass 2: re-run FETCH_HTTP_NODE
        result, message, actions = run_task(state, actions[2])
        self.assertTrue(result)

        result, message, actions = run_task(state, actions[1])  # INTAKE_JSON
        self.assertTrue(result)

        result, message, actions = run_task(state, actions[1])  # JSONLD_COMPACT_DATA
        self.assertTrue(result)
        self.assertEqual(len(actions), 2)
        self.assertEqual(actions[0]['type'], ADD_NODE)
        self.assertEqual(actions[1]['name'], DETECT_AND_VALIDATE_NODE_CLASS)

    @responses.activate
    def test_verify_with_redirection(self):
        url = 'https://example.org/beths-robotics-badge.json'
        assertion_data = json.loads(test_components['2_0_basic_assertion'])
        alt_badge_url = 'http://example.org/altbadgeurl'
        assertion_data['badge'] = alt_badge_url
        responses.add(
            responses.GET, url, json=assertion_data, status=200,
            content_type='application/ld+json'
        )
        set_up_image_mock('https://example.org/beths-robot-badge.png')
        responses.add(
            responses.GET, 'https://w3id.org/openbadges/v2',
            body=test_components['openbadges_context'], status=200,
            content_type='application/ld+json'
        )
        responses.add(
            responses.GET, 'https://example.org/robotics-badge.json',
            body=test_components['2_0_basic_badgeclass'], status=200,
            content_type='application/ld+json'
        )
        responses.add(
            responses.GET, alt_badge_url,
            body=test_components['2_0_basic_badgeclass'], status=200,
            content_type='application/ld+json'
        )
        set_up_image_mock(u'https://example.org/robotics-badge.png')
        responses.add(
            responses.GET, 'https://example.org/organization.json',
            body=test_components['2_0_basic_issuer'], status=200,
            content_type='application/ld+json'
        )

        results = verify(url)

        self.assertEqual(len(results['report']['messages']), 1,
                         "There should be a warning about the redirection.")
        self.assertIn('Node fetched from source', results['report']['messages'][0]['result'],
                      "Message should be the one about the graph ID change.")
        self.assertEqual(len(results['graph']), 3)
        assertion_node = [n for n in results['graph'] if n['id'] == assertion_data['id']][0]
        self.assertEqual(assertion_node['badge'], 'https://example.org/robotics-badge.json')

    @responses.activate
    def test_verify_redirected_validation_subject(self):
        url = 'https://example.org/beths-robotics-badge.json'
        assertion_data = json.loads(test_components['2_0_basic_assertion'])
        alt_assertion_url = 'http://example.org/altbadgeurl'
        assertion_data['id'] = alt_assertion_url
        responses.add(
            responses.GET, url, json=assertion_data, status=200,
            content_type='application/ld+json'
        )
        responses.add(
            responses.GET, alt_assertion_url,
            json=assertion_data, status=200,
            content_type='application/ld+json'
        )
        set_up_image_mock('https://example.org/beths-robot-badge.png')
        responses.add(
            responses.GET, 'https://w3id.org/openbadges/v2',
            body=test_components['openbadges_context'], status=200,
            content_type='application/ld+json'
        )
        responses.add(
            responses.GET, 'https://example.org/robotics-badge.json',
            body=test_components['2_0_basic_badgeclass'], status=200,
            content_type='application/ld+json'
        )
        set_up_image_mock(u'https://example.org/robotics-badge.png')
        responses.add(
            responses.GET, 'https://example.org/organization.json',
            body=test_components['2_0_basic_issuer'], status=200,
            content_type='application/ld+json'
        )

        results = verify(url)

        self.assertEqual(len(results['report']['messages']), 1,
                         "There should be a warning about the redirection.")
        self.assertIn('Node fetched from source', results['report']['messages'][0]['result'],
                      "Message should be the one about the graph ID change.")
        self.assertEqual(len(results['graph']), 3)
        assertion_node = [n for n in results['graph'] if n['id'] == assertion_data['id']][0]
        self.assertEqual(assertion_node['badge'], 'https://example.org/robotics-badge.json')
        self.assertEqual(results['report']['validationSubject'], alt_assertion_url)



OPENBADGES_CONTEXT_V2_DICT = {
  "@context": {
    "id": "@id",
    "type": "@type",

    "extensions": "https://w3id.org/openbadges/extensions#",
    "obi": "https://w3id.org/openbadges#",
    "validation": "obi:validation",

    "cred": "https://w3id.org/credentials#",
    "dc": "http://purl.org/dc/terms/",
    "schema": "http://schema.org/",
    "sec": "https://w3id.org/security#",
    "xsd": "http://www.w3.org/2001/XMLSchema#",

    "AlignmentObject": "schema:AlignmentObject",
    "CryptographicKey": "sec:Key",
    "Endorsement": "cred:Credential",

    "Assertion": "obi:Assertion",
    "BadgeClass": "obi:BadgeClass",
    "Criteria": "obi:Criteria",
    "Evidence": "obi:Evidence",
    "Extension": "obi:Extension",
    "FrameValidation": "obi:FrameValidation",
    "IdentityObject": "obi:IdentityObject",
    "Image": "obi:Image",
    "HostedBadge": "obi:HostedBadge",
    "hosted": "obi:HostedBadge",
    "Issuer": "obi:Issuer",
    "Profile": "obi:Profile",
    "RevocationList": "obi:RevocationList",
    "SignedBadge": "obi:SignedBadge",
    "signed": "obi:SignedBadge",
    "TypeValidation": "obi:TypeValidation",
    "VerificationObject": "obi:VerificationObject",

    "author": { "@id": "schema:author", "@type": "@id" },
    "caption": { "@id": "schema:caption" },
    "claim": {"@id": "cred:claim", "@type": "@id"},
    "created": { "@id": "dc:created", "@type": "xsd:dateTime" },
    "creator": { "@id": "dc:creator", "@type": "@id" },
    "description": { "@id": "schema:description" },
    "email": { "@id": "schema:email" },
    "endorsement": {"@id": "cred:credential", "@type": "@id"},
    "expires": { "@id": "sec:expiration", "@type": "xsd:dateTime" },
    "genre": { "@id": "schema:genre" },
    "image": { "@id": "schema:image", "@type": "@id" },
    "name": { "@id": "schema:name" },
    "owner": {"@id": "sec:owner", "@type": "@id"},
    "publicKey": { "@id": "sec:publicKey", "@type": "@id" },
    "publicKeyPem": { "@id": "sec:publicKeyPem" },
    "related": { "@id": "dc:relation", "@type": "@id" },
    "startsWith": { "@id": "http://purl.org/dqm-vocabulary/v1/dqm#startsWith" },
    "tags": { "@id": "schema:keywords" },
    "targetDescription": { "@id": "schema:targetDescription" },
    "targetFramework": { "@id": "schema:targetFramework" },
    "targetName": { "@id": "schema:targetName" },
    "targetUrl": { "@id": "schema:targetUrl" },
    "telephone": { "@id": "schema:telephone" },
    "url": { "@id": "schema:url", "@type": "@id" },
    "version": { "@id": "schema:version" },

    "alignment": { "@id": "obi:alignment", "@type": "@id" },
    "allowedOrigins": { "@id": "obi:allowedOrigins" },
    "audience": { "@id": "obi:audience" },
    "badge": { "@id": "obi:badge", "@type": "@id" },
    "criteria": { "@id": "obi:criteria", "@type": "@id" },
    "endorsementComment": { "@id": "obi:endorsementComment" },
    "evidence": { "@id": "obi:evidence", "@type": "@id" },
    "hashed": { "@id": "obi:hashed", "@type": "xsd:boolean" },
    "identity": { "@id": "obi:identityHash" },
    "issuedOn": { "@id": "obi:issueDate", "@type": "xsd:dateTime" },
    "issuer": { "@id": "obi:issuer", "@type": "@id" },
    "narrative": { "@id": "obi:narrative" },
    "recipient": { "@id": "obi:recipient", "@type": "@id" },
    "revocationList": { "@id": "obi:revocationList", "@type": "@id" },
    "revocationReason": { "@id": "obi:revocationReason" },
    "revoked": { "@id": "obi:revoked", "@type": "xsd:boolean" },
    "revokedAssertions": { "@id": "obi:revoked" },
    "salt": { "@id": "obi:salt" },
    "targetCode": { "@id": "obi:targetCode" },
    "uid": { "@id": "obi:uid" },
    "validatesType": "obi:validatesType",
    "validationFrame": "obi:validationFrame",
    "validationSchema": "obi:validationSchema",
    "verification": { "@id": "obi:verify", "@type": "@id" },
    "verificationProperty": { "@id": "obi:verificationProperty" },
    "verify": "verification"
  }
}

OPENBADGES_CONTEXT_V1_URI = "https://w3id.org/openbadges/v1"
OPENBADGES_CONTEXT_V2_URI = "https://w3id.org/openbadges/v2"



"""
INPUT Actions:
Store and process user input via the API
"""
SET_INPUT_TYPE = 'SET_INPUT_TYPE'
SET_PROFILE_ID = 'SET_PROFILE_ID'
STORE_INPUT = 'STORE_INPUT'
STORE_ORIGINAL_RESOURCE = 'STORE_ORIGINAL_JSON'

"""
GRAPH Actions:
Manage the state of the known entities related to the validation subject.
"""
ADD_NODE = 'ADD_NODE'
PATCH_NODE = 'PATCH_NODE'
UPDATE_NODE = 'UPDATE_NODE'
PATCH_NODE_REFERENCE = 'PATCH_NODE_REFERENCE'

"""
TASK Actions:
Add, update, and complete tasks.
"""
ADD_TASK = 'ADD_TASK'
UPDATE_TASK = 'UPDATE_TASK'
DELETE_TASK = 'DELETE_TASK'
DELETE_OUTDATED_NODE_TASKS = 'DELETE_OUTDATED_NODE_TASKS'
REPORT_MESSAGE = 'REPORT_MESSAGE'
RESOLVE_TASK = 'RESOLVE_TASK'
TRIGGER_CONDITION = 'TRIGGER_CONDITION'


"""
VALIDATION REPORT Actions:
Note data that should be stored in the validation report.
"""
RUN_VALIDATION_REPORT = 'RUN_VALIDATION_REPORT'
SET_OPENBADGES_VERSION = 'SET_OPENBADGES_VERSION'
SET_VALIDATION_SUBJECT = 'SET_VALIDATION_SUBJECT'
SET_VERIFIED_PROFILE = 'SET_VERIFIED_PROFILE'



import responses
import unittest

from openbadges.verifier.actions.tasks import add_task
from openbadges.verifier.openbadges_context import OPENBADGES_CONTEXT_V2_URI
from openbadges.verifier.tasks.task_types import VALIDATE_EXPECTED_NODE_CLASS
from openbadges.verifier.tasks.validation import OBClasses
from openbadges import verify
from openbadges.verifier.tasks import task_named


from .utils import set_up_context_mock, set_up_image_mock


class EndorsementTests(unittest.TestCase):
    def set_up_resources(self):
        self.assertion = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'type': 'Assertion',
            'id': 'http://example.com/assertion',
            'badge': 'http://example.com/badgeclass',
            'issuedOn': '2017-09-30T00:00Z',
            'recipient': {'identity': 'recipient@example.com', 'type': 'email', 'hashed': False},
            'verification': {
                'type': 'HostedBadge'
            }
        }
        self.badgeclass = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': 'http://example.com/badgeclass',
            'type': 'BadgeClass',
            'issuer': 'http://example.com/issuer',
            'endorsement': {
                '@context': OPENBADGES_CONTEXT_V2_URI,
                'id': 'http://example.org/endorsement',
                'type': 'Endorsement',
                'claim': {
                    'id': 'http://example.com/badgeclass',
                    'endorsementComment': 'Pretty good'
                },
                'issuedOn': '2017-10-01T00:00Z',
                'issuer': 'http://example.org/issuer',
                'verification': {
                    'type': "HostedBadge"
                }
            },
            'name': 'Best Badge',
            'description': 'An achievement that is good.',
            'image': 'http://example.com/badgeimage',
            'criteria': 'http://example.com/badgecriteria'
        }
        self.badgeclass_with_endorsement_array = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': 'http://example.com/badgeclass',
            'type': 'BadgeClass',
            'issuer': 'http://example.com/issuer',
            'endorsement': ['http://example.org/endorsement', 'http://example.org/endorsement2'],
            'name': 'Best Badge',
            'description': 'An achievement that is good.',
            'image': 'http://example.com/badgeimage',
            'criteria': 'http://example.com/badgecriteria'
        }
        self.issuer = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': 'http://example.com/issuer',
            'type': 'Profile',
            'name': 'Test Issuer',
            'email': 'test@example.com',
            'url': 'http://example.com'
        }
        self.endorsement = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': 'http://example.org/endorsement',
            'type': 'Endorsement',
            'claim': {
                'id': self.badgeclass['id'],
                'endorsementComment': 'Pretty good'
            },
            'issuedOn': '2017-10-01T00:00Z',
            'issuer': 'http://example.org/issuer',
            'verification': {
                'type': "HostedBadge"
            }
        }
        self.endorsement2 = self.endorsement.copy()
        self.endorsement2['id'] = 'http://example.org/endorsement2'
        self.endorsement_issuer = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': 'http://example.org/issuer',
            'type': 'Profile',
            'name': 'Test Endorser',
            'email': 'test@example.org',
            'url': 'http://example.org'
        }

    @responses.activate
    def test_validate_linked_endorsement(self):
        set_up_context_mock()
        self.set_up_resources()

        for resource in [self.assertion, self.badgeclass, self.issuer, self.endorsement, self.endorsement_issuer]:
            responses.add(responses.GET, resource['id'], json=resource)
        set_up_image_mock(self.badgeclass['image'])

        results = verify(self.assertion['id'])
        self.assertTrue(results['report']['valid'])
        self.assertEqual(5, len(results['graph']), "The graph now contains all five resources.")

    @responses.activate
    def test_validate_linked_endorsement_array(self):
        set_up_context_mock()
        self.set_up_resources()

        for resource in [self.assertion, self.badgeclass_with_endorsement_array, self.issuer, self.endorsement, self.endorsement2, self.endorsement_issuer]:
            responses.add(responses.GET, resource['id'], json=resource)
        set_up_image_mock(self.badgeclass['image'])

        results = verify(self.assertion['id'])
        self.assertTrue(results['report']['valid'])
        self.assertEqual(6, len(results['graph']), "The graph now contains all six resources including both endorsements.")

    @responses.activate
    def test_validate_endorsement_as_input(self):
        set_up_context_mock()
        self.set_up_resources()

        for resource in [self.assertion, self.badgeclass, self.issuer, self.endorsement, self.endorsement_issuer]:
            responses.add(responses.GET, resource['id'], json=resource)
        set_up_image_mock(self.badgeclass['image'])

        results = verify(self.endorsement['id'])
        self.assertTrue(results['report']['valid'])
        self.assertEqual(2, len(results['graph']))

    def test_claim_property_validation(self):
        self.set_up_resources()

        options = {'max_validation_depth': 3}

        state = {'graph': [self.endorsement]}
        task_meta = add_task(
            VALIDATE_EXPECTED_NODE_CLASS, node_id=self.endorsement['id'], prop_name='claim',
            expected_class=OBClasses.Endorsement,
            depth=0
        )

        result, message, actions = task_named(VALIDATE_EXPECTED_NODE_CLASS)(state, task_meta, **options)
        self.assertTrue(result)
        claim_action = [a for a in actions if a.get('prop_name') == 'claim'][0]

        result, message, actions = task_named(claim_action['name'])(state, claim_action, **options)
        self.assertTrue(result)
        self.assertEqual(1, len(actions))

        result, message, actions = task_named(actions[0]['name'])(state, actions[0], **options)
        self.assertTrue(result)
        self.assertEqual(3, len(actions))



import responses
import unittest

from pydux import create_store

from openbadges.verifier import verify
from openbadges.verifier.reducers import main_reducer
from openbadges.verifier.state import (filter_active_tasks, INITIAL_STATE, get_node_by_id,
                              get_node_by_path,)

try:
    from .testfiles.test_components import test_components
except (ImportError, SystemError):
    from .testfiles.test_components import test_components


class InitializationTests(unittest.TestCase):
    def test_store_initialization(self):
        def no_op(state, action):
            return state
        store = create_store(no_op, INITIAL_STATE)
        self.assertEqual(store.get_state(), INITIAL_STATE)


class TaskFilterTests(unittest.TestCase):
    def test_active_filter(self):
        tasks = [
            {'name': 'Carl', 'complete': True},
            {'name': 'Tim', 'complete': True},
            {'name': 'Linda', 'complete': False},
        ]
        state = {'tasks': tasks}

        active_tasks = filter_active_tasks(state)
        self.assertEqual(len(active_tasks), 1, "There should only be one active (incomplete) task")
        self.assertEqual(active_tasks[0]['name'], 'Linda')

        mary = {'name': 'Mary', 'complete': False, 'prerequisites': 'Tim'}
        tasks.append(mary)
        active_tasks = filter_active_tasks(state)
        self.assertEqual(len(active_tasks), 2, "Task whose prereq has been met should be active")

        mary['prerequisites'] = ['Tim', 'Carl']
        active_tasks = filter_active_tasks(state)
        self.assertEqual(len(active_tasks), 2, "Task whose prereqs has been met should be active")

        mary['prerequisites'] = ['Tim', 'Linda']
        active_tasks = filter_active_tasks(state)
        self.assertEqual(len(active_tasks), 1, "Task with an incomplete prereq should not be active")


class FindNodeByPathTests(unittest.TestCase):
    def test_find_node_with_single_length_path(self):
        state = {
            'graph': [{'id': '_:b0'}]
        }
        with self.assertRaises(IndexError):
            get_node_by_path(state, ['_:b100'])

        self.assertEqual(get_node_by_path(state, ['_:b0']), state['graph'][0])

        state['graph'].append({'id': '_:b1', 'prop': '_:b0'})
        self.assertEqual(get_node_by_path(state, ['_:b1', 'prop']), state['graph'][0])

        state['graph'].append({'id': '_:b2', 'prop': ['http://unknown.external', '_:b1']})
        self.assertEqual(get_node_by_path(state, ['_:b2', 'prop', 1]), state['graph'][1])
        self.assertEqual(get_node_by_path(state, ['_:b2', 'prop', 1, 'prop']), state['graph'][0])



from pyld import jsonld
import re
import rfc3986
import six

from ..openbadges_context import OPENBADGES_CONTEXT_V2_URI

URN_REGEX = r'^urn:uuid:[0-9a-f]{8}-[0-9a-f]{4}-[1-5][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}$'


def task_result(success=True, message='', actions=None):
    """
    Formats a response from a task so that the task caller can dispatch actions
    :param success: bool
    :param message: str
    :param actions: list(dict)
    :return: tuple
    """
    if not actions:
        actions = []

    return (success, message, actions,)


def is_empty_list(value):
    return isinstance(value, (tuple, list)) and len(value) == 0


def is_null_list(value):
    return isinstance(value, (tuple, list)) and all(val is None for val in value)


def abbreviate_value(value, length=48):
    if isinstance(value, (tuple, list)):
        value = ', '.join([str(val) for val in value])

    if len(str(value)) < length:
        return str(value)
    else:
        return str(value)[:length] + '...'


def abbreviate_node_id(node_id=None, node_path=None, length=48):
    if node_id:
        return node_id
    return abbreviate_value(node_path, length=length)


def is_iri(value):
    return bool(
        is_url(value) or
        re.match(r'^_:', value) or
        re.match(URN_REGEX, value, re.IGNORECASE)
    )


def is_blank_node_id(value):
    return bool(re.match(r'^_:', value))


def is_data_uri(value):
    data_uri_regex = r'(?P<scheme>^data):(?P<mimetypes>[^,]{0,}?)?(?P<encoding>base64)?,(?P<data>.*$)'
    ret = False
    try:
        if ((value and isinstance(value, six.string_types))
                and rfc3986.is_valid_uri(value, require_scheme=True)
                and re.match(data_uri_regex, value, re.IGNORECASE)
                and re.match(data_uri_regex, value, re.IGNORECASE).group('scheme').lower() == 'data'):
            ret = True
    except ValueError:
        pass
    return ret


def is_url(value):
    ret = False
    try:
        if ((value and isinstance(value, six.string_types))
                and rfc3986.is_valid_uri(value, require_scheme=True)
                and rfc3986.uri_reference(value).scheme.lower() in ['http', 'https']):
            ret = True
    except ValueError:
        pass
    return ret


def filter_tasks(state, **kwargs):
    tasks = state.get('tasks', [])

    def _matches(val):
        return all([val.get(kwarg) == kwargs[kwarg] for kwarg in list(kwargs.keys())])

    return list(filter(_matches, tasks))


def is_ld_term_in_list(term, search_list, context=OPENBADGES_CONTEXT_V2_URI, options=None):
    construct = {
        '@context': context,
        '_:term': {'@type': term},
        '_:list': {'@type': search_list}
    }
    result = jsonld.expand(construct, options)
    return result[0]['_:term'][0]['@type'][0] in result[0]['_:list'][0]['@type']


def combine_contexts(*args):
    if not len(args):
        return []

    context = args[0]
    if context is None:
        return [] + combine_contexts(*args[1:])
    elif isinstance(context, dict):
        try:
            return combine_contexts(context['@context']) + combine_contexts(*args[1:])
        except KeyError:
            return [context] + combine_contexts(*args[1:])
    elif isinstance(context, six.string_types):
        return [context] + combine_contexts(*args[1:])
    elif isinstance(context, (list, tuple)):
        return context + combine_contexts(*args[1:])



# coding=utf-8
from datetime import datetime, timedelta
import json
from pyld import jsonld
from pytz import utc
from pydux import create_store
import responses
import unittest

from openbadges.verifier.actions.action_types import ADD_TASK, PATCH_NODE
from openbadges.verifier.actions.graph import add_node, patch_node
from openbadges.verifier.actions.tasks import add_task
from openbadges.verifier.openbadges_context import OPENBADGES_CONTEXT_V2_DICT, OPENBADGES_CONTEXT_V2_URI
from openbadges.verifier.reducers import main_reducer
from openbadges.verifier.state import filter_active_tasks, INITIAL_STATE
from openbadges.verifier.tasks import task_named, run_task
from openbadges.verifier.tasks.validation import (_get_validation_actions, assertion_timestamp_checks,
                                         criteria_property_dependencies, detect_and_validate_node_class,
                                         OBClasses, PrimitiveValueValidator, validate_property, ValueTypes,)
from openbadges.verifier.tasks.verification import (_default_verification_policy, hosted_id_in_verification_scope,)
from openbadges.verifier.tasks.task_types import (ASSERTION_TIMESTAMP_CHECKS, CRITERIA_PROPERTY_DEPENDENCIES,
                                         DETECT_AND_VALIDATE_NODE_CLASS, HOSTED_ID_IN_VERIFICATION_SCOPE,
                                         IDENTITY_OBJECT_PROPERTY_DEPENDENCIES, ISSUER_PROPERTY_DEPENDENCIES,
                                         VALIDATE_RDF_TYPE_PROPERTY, VALIDATE_PROPERTY, VALIDATE_EXPECTED_NODE_CLASS)
from openbadges.verifier.utils import MESSAGE_LEVEL_WARNING
from openbadges.verifier.verifier import call_task, verify

try:
    from .testfiles.test_components import test_components
    from tests.utils import set_up_context_mock, set_up_image_mock
except (ImportError, SystemError):
    from .testfiles.test_components import test_components
    from tests.utils import set_up_context_mock, set_up_image_mock


class PropertyValidationTests(unittest.TestCase):

    def test_data_uri_validation(self):
        validator = PrimitiveValueValidator(ValueTypes.DATA_URI)
        good_uris = ('data:image/gif;base64,R0lGODlhyAAiALM...DfD0QAADs=',
                     'data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg==',
                     'data:text/plain;charset=UTF-8;page=21,the%20data:1234,5678',
                     'data:text/vnd-example+xyz;foo=bar;base64,R0lGODdh',
                     'data:,actually%20a%20valid%20data%20URI',
                     'data:,')
        bad_uris = ('data:image/gif',
                    'http://someexample.org',
                    'data:bad:path')
        for uri in good_uris:
            self.assertTrue(validator(uri), "`{}` should pass data URI validation but failed.".format(uri))
        for uri in bad_uris:
            self.assertFalse(validator(uri), "`{}` should fail data URI/URL validation but passed.".format(uri))

    def test_data_uri_or_url_validation(self):
        validator = PrimitiveValueValidator(ValueTypes.DATA_URI_OR_URL)
        good_uris = ('data:image/gif;base64,R0lGODlhyAAiALM...DfD0QAADs=',
                     'data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg==',
                     'data:text/plain;charset=UTF-8;page=21,the%20data:1234,5678',
                     'data:text/vnd-example+xyz;foo=bar;base64,R0lGODdh',
                     'http://www.example.com:8080/', 'http://www.example.com:8080/foo/bar',
                     'http://www.example.com/foo%20bar', 'http://www.example.com/foo/bar?a=b&c=d',
                     'http://www.example.com/foO/BaR', 'HTTPS://www.EXAMPLE.cOm/',
                     'http://142.42.1.1:8080/', 'http://142.42.1.1/',
                     'http://foo.com/blah_(wikipedia)#cite-1', 'http://a.b-c.de',
                     'http://userid:password@example.com/', "http://-.~:%40:80%2f:password@example.com",
                     'http://code.google.com/events/#&product=browser')
        bad_uris = ('///', '///f', '//',
                    'rdar://12345', 'h://test', 'http:// shouldfail.com', ':// should fail', '', 'a',
                    'urn:uuid:129487129874982374', 'urn:uuid:9d278beb-36cf-4bc8-888d-674ff9843d72')
        for uri in good_uris:
            self.assertTrue(validator(uri), "`{}` should pass data URI/URL validation but failed.".format(uri))
        for uri in bad_uris:
            self.assertFalse(validator(uri), "`{}` should fail data URI/URL validation but passed.".format(uri))

    def test_url_validation(self):
        validator = PrimitiveValueValidator(ValueTypes.URL)
        # Thanks to Mathias Bynens for fun URL examples: http://mathiasbynens.be/demo/url-regex
        good_urls = ('http://www.example.com:8080/', 'http://www.example.com:8080/foo/bar',
                     'http://www.example.com/foo%20bar', 'http://www.example.com/foo/bar?a=b&c=d',
                     'http://www.example.com/foO/BaR', 'HTTPS://www.EXAMPLE.cOm/',
                     'http://142.42.1.1:8080/', 'http://142.42.1.1/', 'http://localhost:3000/123',
                     'http://foo.com/blah_(wikipedia)#cite-1', 'http://a.b-c.de',
                     'http://userid:password@example.com/', "http://-.~:%40:80%2f:password@example.com",
                     'http://code.google.com/events/#&product=browser')
        good_urls_that_fail = ('http://✪df.ws/123', 'http://عمان.icom.museum/',)  # TODO: Discuss support for these
        bad_urls = ('data:image/gif;base64,R0lGODlhyAAiALM...DfD0QAADs=', '///', '///f', '//',
                    'rdar://12345', 'h://test', 'http:// shouldfail.com', ':// should fail', '', 'a',
                    'urn:uuid:129487129874982374', 'urn:uuid:9d278beb-36cf-4bc8-888d-674ff9843d72')
        bad_urls_that_pass = ('http://', 'http://../', 'http://foo.bar?q=Spaces should be encoded',
                                          'http://f', 'http://-error-.invalid/', 'http://.www.foo.bar./',)

        for url in good_urls:
            self.assertTrue(validator(url), "`{}` should pass URL validation but failed.".format(url))
        for url in bad_urls:
            self.assertFalse(validator(url), "`{}` should fail URL validation but passed.".format(url))

    def test_iri_validation(self):
        validator = PrimitiveValueValidator(ValueTypes.IRI)
        # Thanks to Mathias Bynens for fun URL examples: http://mathiasbynens.be/demo/url-regex
        good_iris = ('http://www.example.com:8080/', '_:b0', '_:b12', '_:b107', '_:b100000001232',
                     'urn:uuid:9d278beb-36cf-4bc8-888d-674ff9843d72',
                     'urn:uuid:9D278beb-36cf-4bc8-888d-674ff9843d72'
                     )
        good_iris_that_fail = ()  # TODO: Discuss support for these
        bad_iris = ('data:image/gif;base64,R0lGODlhyAAiALM...DfD0QAADs=', 'urn:uuid', 'urn:uuid:123',
                    '', 'urn:uuid:', 'urn:uuid:zz278beb-36cf-4bc8-888d-674ff9843d72',)
        bad_iris_that_pass = ()

        for url in good_iris:
            self.assertTrue(validator(url), "`{}` should pass IRI validation but failed.".format(url))
        for url in bad_iris:
            self.assertFalse(validator(url), "`{}` should fail IRI validation but passed.".format(url))

    def test_url_authority(self):
        validator = PrimitiveValueValidator(ValueTypes.URL_AUTHORITY)
        good_values = (
            'google.com', 'nerds.example.com'
        )
        bad_values = (
            '666', 'http://google.com/', 'https://www.google.com/search?q=murder+she+wrote&oq=murder+she+wrote',
            'ftp://123.123.123.123', 'bears', 'lots of hungry bears', 'bears.com/thewoods',
            '192.168.0.1', '1::6:7:8'
        )

        for value in good_values:
            self.assertTrue(validator(value), "{} should pass origin validation but failed".format(value))
        for value in bad_values:
            self.assertFalse(validator(value), "{} should fail origin validation but passed".format(value))


class IriPropertyValidationTests(unittest.TestCase):
    def test_validate_compacted_iri_value(self):
        first_node = {
            'type': 'Issuer',
            'id': 'https://example.org/issuer1',
            'verification': {
                'verificationProperty': 'id'
            }
        }
        state = {'graph': [first_node]}

        validator = PrimitiveValueValidator(ValueTypes.COMPACT_IRI)
        self.assertTrue(validator('id'))
        self.assertTrue(validator('email'))
        self.assertTrue(validator('telephone'))
        self.assertTrue(validator('url'))
        self.assertFalse(validator('sloths'))

        task = add_task(
            VALIDATE_PROPERTY,
            node_path=[first_node['id'], 'verification'],
            prop_name='verificationProperty',
            prop_type=ValueTypes.COMPACT_IRI,
            required=False
        )

        result, message, actions = validate_property(state, task)
        self.assertTrue(result)


class PropertyValidationTaskTests(unittest.TestCase):

    def test_basic_text_property_validation(self):
        first_node = {'id': 'http://example.com/1', 'string_prop': 'string value'}
        state = {
            'graph': [first_node]
        }
        task = add_task(
            VALIDATE_PROPERTY,
            node_id=first_node['id'],
            prop_name='string_prop',
            prop_type=ValueTypes.TEXT,
            required=False
        )

        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "Optional property is present and correct; validation should pass.")

        task['required'] = True
        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "Required property is present and correct; validation should pass.")

        first_node['string_prop'] = 1
        result, message, actions = validate_property(state, task)
        self.assertFalse(result, "Required string property is an int; validation should fail")
        self.assertEqual(
            message, "TEXT property string_prop value 1 not valid in unknown type node {}".format(first_node['id'])
        )

        task['required'] = False
        result, message, actions = validate_property(state, task)
        self.assertFalse(result, "Optional string property is an int; validation should fail")
        self.assertEqual(
            message, "TEXT property string_prop value 1 not valid in unknown type node {}".format(first_node['id'])
        )

        # When property isn't present
        second_node = {'id': 'http://example.com/1'}
        state = {'graph': [second_node]}
        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "Optional property is not present; validation should pass.")

        task['required'] = True
        result, message, actions = validate_property(state, task)
        self.assertFalse(result, "Required property is not present; validation should fail.")

    def test_basic_telephone_property_validation(self):
        first_node = {
            'id': 'http://example.com'
        }
        state = {
            'graph': [first_node]
        }
        task = add_task(
            VALIDATE_PROPERTY,
            node_id=first_node['id'],
            prop_name='tel',
            required=True,
            prop_type=ValueTypes.TELEPHONE
        )

        good_values = ["+64010", "+15417522845", "+18006664358", "+18006662344;ext=666"]
        bad_values = ["1-800-666-DEVIL", "1 (555) 555-5555", "+99 55 22 1234", "+18006664343 x666"]

        for tel_value in good_values:
            first_node['tel'] = tel_value
            result, message, actions = validate_property(state, task)
            self.assertTrue(result)

        for tel_value in bad_values:
            first_node['tel'] = tel_value
            result, message, actions = validate_property(state, task)
            self.assertFalse(result)

    def test_basic_email_property_validation(self):
        first_node = {
            'id': 'http://example.com'
        }
        state = {
            'graph': [first_node]
        }
        task = add_task(
            VALIDATE_PROPERTY,
            node_id=first_node['id'],
            prop_name='email',
            required=True,
            prop_type=ValueTypes.EMAIL
        )

        good_values = ["abc@localhost", "cool+uncool@example.org"]
        bad_values = [" spacey@gmail.com", "steveman [at] gee mail dot com"]

        for val in good_values:
            first_node['email'] = val
            result, message, actions = validate_property(state, task)
            self.assertTrue(result, "{} should be marked a valid email".format(val))

        for val in bad_values:
            first_node['email'] = val
            result, message, actions = validate_property(state, task)
            self.assertFalse(result, "{} should be marked an invalid email".format(val))

    def test_basic_boolean_property_validation(self):
        first_node = {'id': 'http://example.com/1'}
        state = {
            'graph': [first_node]
        }
        task = add_task(
            VALIDATE_PROPERTY,
            node_id=first_node['id'],
            prop_name='bool_prop',
            required=False,
            prop_type=ValueTypes.BOOLEAN
        )

        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "Optional property is not present; validation should pass.")
        self.assertEqual(
            message, "Optional property bool_prop not present in unknown type node {}".format(first_node['id'])
        )

        task['required'] = True
        result, message, actions = validate_property(state, task)
        self.assertFalse(result, "Required property is not present; validation should fail.")
        self.assertEqual(
            message, "Required property bool_prop not present in unknown type node {}".format(first_node['id'])
        )

        first_node['bool_prop'] = True
        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "Required boolean property matches expectation")
        self.assertEqual(
            message, "BOOLEAN property bool_prop is valid in unknown type node {}".format(first_node['id'])
        )

    def test_basic_id_validation(self):
        ## test assumes that a node's ID must be an IRI
        first_node = {'id': 'http://example.com/1'}
        state = {
            'graph': [first_node]
        }
        task = add_task(
            VALIDATE_PROPERTY,
            node_id=first_node['id'],
            prop_name='id',
            required=True,
            prop_type=ValueTypes.IRI
        )

        result, message, actions = validate_property(state, task)
        self.assertTrue(result)
        self.assertEqual(len(actions), 0)

    def test_basic_image_prop_validation(self):
        _VALID_DATA_URI = 'data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg=='
        _VALID_IMAGE_URL = 'http://example.com/images/foo.png'
        _INVALID_URI = 'notanurl'
        first_node = {'id': 'http://example.com/1',
                      'image_prop': _VALID_DATA_URI}

        state = {
            'graph': [first_node]
        }

        task = add_task(
            VALIDATE_PROPERTY,
            node_id=first_node['id'],
            prop_name='image_prop',
            required=False,
            prop_type=ValueTypes.DATA_URI_OR_URL
        )

        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "Optional image_prop URI is present and well-formed; validation should pass.")
        self.assertTrue(
            "DATA_URI_OR_URL" in message and "valid" in message
        )

        first_node['image_prop'] = _INVALID_URI
        result, message, actions = validate_property(state, task)
        self.assertFalse(result, "Optional image prop is present and mal-formed; validation should fail.")
        self.assertEqual(
            message, "DATA_URI_OR_URL property image_prop value {} not valid in unknown type node {}".format(
                _INVALID_URI,
                first_node['id'])
        )

    def test_basic_url_prop_validation(self):
        _VALID_URL = 'http://example.com/2'
        _INVALID_URL = 'notanurl'
        first_node = {'id': 'http://example.com/1',
                      'url_prop': _VALID_URL}
        state = {
            'graph': [first_node]
        }

        task = add_task(
            VALIDATE_PROPERTY,
            node_id=first_node['id'],
            prop_name='url_prop',
            required=False,
            prop_type=ValueTypes.URL
        )

        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "Optional URL prop is present and well-formed; validation should pass.")

        first_node['url_prop'] = _INVALID_URL
        result, message, actions = validate_property(state, task)
        self.assertFalse(result, "Optional URL prop is present and mal-formed; validation should fail.")
        self.assertEqual(
            message, "URL property url_prop value {} not valid in unknown type node {}".format(
                _INVALID_URL,
                first_node['id'])
        )

    def test_basic_datetime_property_validation(self):
        _VALID_DATETIMES = ['1977-06-10T12:00:00+0800',
                            '1977-06-10T12:00:00-0800',
                            '1977-06-10T12:00:00+08',
                            '1977-06-10T12:00:00+08:00']
        _INVALID_NOTZ_DATETIMES = ['1977-06-10T12:00:00']
        _INVALID_DATETIMES = ['notadatetime']

        first_node = {'id': 'http://example.com/1', 'date_prop': '1977-06-10T12:00:00Z'}
        state = {
            'graph': [first_node]
        }
        task = add_task(
            VALIDATE_PROPERTY,
            node_id=first_node['id'],
            prop_name='date_prop',
            required=False,
            prop_type=ValueTypes.DATETIME
        )
        task['id'] = 1

        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "Optional date prop is present and well-formed; validation should pass.")
        self.assertEqual(
            message, "DATETIME property date_prop is valid in unknown type node {}".format(first_node['id'])
        )

        for date in _VALID_DATETIMES:
            first_node['date_prop'] = date
            result, message, actions = validate_property(state, task)
            self.assertTrue(result,
                            "Optional date prop {} is well-formed; validation should pass.".format(date))

        for date in _INVALID_NOTZ_DATETIMES:
            first_node['date_prop'] = date
            result, message, actions = validate_property(state, task)
            self.assertFalse(result, "Optional date prop has no tzinfo particle; validation should fail.")
            self.assertEqual(
                message,
                "DATETIME property date_prop value {} not valid in unknown type node {}".format(
                    str(date), first_node['id'])
            )

        for date in _INVALID_DATETIMES:
            first_node['date_prop'] = date
            result, message, actions = validate_property(state, task)
            self.assertFalse(result, "Optional date prop has malformed datetime; validation should fail.")
            self.assertEqual(
                message,
                "DATETIME property date_prop value {} not valid in unknown type node {}".format(
                    str(date), first_node['id'])
            )

    def test_validation_action(self):
        store = create_store(main_reducer, INITIAL_STATE)
        first_node = {
            'text_prop': 'text_value',
            'bool_prop': True
        }
        store.dispatch(add_node(node_id="http://example.com/1", data=first_node))

        # 0. Test of an existing valid text prop: expected pass
        store.dispatch(add_task(
            VALIDATE_PROPERTY,
            node_id="http://example.com/1",
            prop_name="text_prop",
            required=True,
            prop_type=ValueTypes.TEXT
        ))

        # 1. Test of an missing optional text prop: expected pass
        store.dispatch(add_task(
            VALIDATE_PROPERTY,
            node_id="http://example.com/1",
            prop_name="nonexistent_text_prop",
            required=False,
            prop_type=ValueTypes.TEXT
        ))

        # 2. Test of an present optional valid boolean prop: expected pass
        store.dispatch(add_task(
            VALIDATE_PROPERTY,
            node_id="http://example.com/1",
            prop_name="bool_prop",
            required=False,
            prop_type=ValueTypes.BOOLEAN
        ))

        num_actions = len(store.get_state()['tasks'])
        # 2b. Retesting the same property won't add any actions to state
        store.dispatch(add_task(
            VALIDATE_PROPERTY,
            node_id="http://example.com/1",
            prop_name="bool_prop",
            required=True,
            prop_type=ValueTypes.TEXT
        ))
        self.assertEqual(num_actions, len(store.get_state()['tasks']),
                         "Duplicate check shouldn't add another action")

        store.dispatch(patch_node(node_id="http://example.com/1", data={'bool_prop2': True}))
        # 3 Invalid text value for required bool prop should fail.
        store.dispatch(add_task(
            VALIDATE_PROPERTY,
            node_id="http://example.com/1",
            prop_name="bool_prop2",
            required=True,
            prop_type=ValueTypes.TEXT
        ))

        # 4. Test of a required missing boolean prop: expected fail
        store.dispatch(add_task(
            VALIDATE_PROPERTY,
            node_id="http://example.com/1",
            prop_name="nonexistent_bool_prop",
            required=True,
            prop_type=ValueTypes.BOOLEAN
        ))

        last_task_id = 0
        while len(filter_active_tasks(store.get_state())):
            active_tasks = filter_active_tasks(store.get_state())
            task_meta = active_tasks[0]
            task_func = task_named(task_meta['name'])

            if task_meta['task_id'] == last_task_id:
                break

            last_task_id = task_meta['task_id']
            call_task(task_func, task_meta, store)

        state = store.get_state()
        self.assertEqual(len(state['tasks']), 5)
        self.assertTrue(state['tasks'][0]['success'], "Valid required text property is present.")
        self.assertTrue(state['tasks'][1]['success'], "Missing optional text property is OK.")
        self.assertTrue(state['tasks'][2]['success'], "Valid optional boolean property is present.")
        self.assertFalse(state['tasks'][3]['success'], "Invalid required text property is present.")
        self.assertFalse(state['tasks'][4]['success'], "Required boolean property is missing.")

    def optional_accepts_null_values(self):
        """
        If a simple or id-type property is optional, null values should not be rejected.
        """
        pass
        # TODO: they should be filtered from the node reported at output.

    def test_many_validation(self):
        """
        When detect_and_validate_node_class (through _get_validation_actions)
        queue up actions, some configs may have many=True. This means single or multiple
        values should be accepted. If optional, empty lists should also be accepted.
        """
        first_node = {
            'id': '_:b0',
            'type': 'BadgeClass'
        }
        state = {'graph': [first_node]}

        values = (
            ['one', 'two', 'three'],  # values[0]
            'one',                    # values[1]
            1,                        # values[2]
            ['one', 2, 'three'],      # values[3]
            [],                       # values[4]
            None,                     # values[5]
            [None],                   # values[6]
            [None, None],             # values[7] Always fail
            [None, 'one', None]       # values[8] Always fail
        )

        task_config = {
            'node_id': first_node['id'], 'node_class': OBClasses.BadgeClass,
            'prop_name': 'tags', 'prop_type': ValueTypes.TEXT
        }

        t_result = t_message = t_actions = None
        def run(i, expect_success, message):
            first_node['tags'] = values[i]
            task = add_task(VALIDATE_PROPERTY, **task_config)
            t_result, t_message, t_actions = validate_property(state, task)
            msg = message + ' when required is {} and many is {}'
            self.assertEqual(t_result, expect_success, msg.format(
                task_config['required'], task_config['many']))

        task_config['required'] = True
        task_config['many'] = True
        run(0, True, "Many acceptable values pass"),
        run(1, True, "One acceptable value passes"),
        run(2, False, "One unacceptable value fails"),
        run(3, False, "List containing an unacceptable non-null value fails"),
        run(4, False, "Empty list fails"),
        run(5, False, "One null value fails"),
        run(6, False, "One null value in a list fails"),
        run(7, False, "A list with all null values fails"),
        run(8, False, "One acceptable value in a list with other null values fails"),
        del first_node['tags']
        task = add_task(VALIDATE_PROPERTY, **task_config)
        t_result, t_message, t_actions = validate_property(state, task)
        self.assertFalse(t_result, "No value present fails when required is True.")

        task_config['required'] = True
        task_config['many'] = False
        run(0, False, "Many acceptable values fail"),
        run(1, True, "One acceptable value passes"),
        run(2, False, "One unacceptable value fails"),
        run(3, False, "List containing an unacceptable non-null value fails"),
        run(4, False, "Empty list fails"),
        run(5, False, "One null value fails"),
        run(6, False, "One null value in a list fails"),
        run(7, False, "Multiple null values fail"),
        run(8, False, "One acceptable value in a list with other null values fails"),
        del first_node['tags']
        task = add_task(VALIDATE_PROPERTY, **task_config)
        t_result, t_message, t_actions = validate_property(state, task)
        self.assertFalse(t_result, "No value present fails when required is True.")

        task_config['required'] = False
        task_config['many'] = True
        run(0, True, "Many acceptable values pass"),
        run(1, True, "One acceptable value passes"),
        run(2, False, "One unacceptable value fails"),
        run(3, False, "List containing an unacceptable non-null value fails"),
        run(4, True, "Empty list passes"),
        run(5, True, "One null value passes"),
        run(6, True, "One null value in a list passes"),
        run(7, True, "Multiple null values pass"),
        run(8, False, "One acceptable value in a list with other null values fails"),
        del first_node['tags']
        task = add_task(VALIDATE_PROPERTY, **task_config)
        t_result, t_message, t_actions = validate_property(state, task)
        self.assertTrue(t_result, "No value present fails when required is False.")

        task_config['required'] = False
        task_config['many'] = False
        run(0, False, "Many acceptable values fail"),
        run(1, True, "One acceptable value passes"),
        run(2, False, "One unacceptable value fails"),
        run(3, False, "List containing an unacceptable non-null value fails"),
        run(4, True, "Empty list passes"),
        run(5, True, "One null value passes"),
        run(6, True, "One null value in a list passes"),
        run(7, True, "Multiple null values pass"),
        run(8, False, "One acceptable value in a list with other null values fails"),
        del first_node['tags']
        task = add_task(VALIDATE_PROPERTY, **task_config)
        t_result, t_message, t_actions = validate_property(state, task)
        self.assertTrue(t_result, "No value present fails when required is False.")


class IDPropertyValidationTests(unittest.TestCase):
    def test_validate_nested_identity_object(self):
        first_node = {
            'id': 'http://example.com/1',
            'recipient': {
                'id': '_:b0',
                'identity': 'two@example.com',
                'type': 'email',
                'hashed': False
            }
        }
        state = {'graph': [first_node]}

        task = add_task(
            VALIDATE_PROPERTY,
            node_id="http://example.com/1",
            prop_name="recipient",
            required=True,
            prop_type=ValueTypes.ID,
            expected_class=OBClasses.IdentityObject
        )

        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "Property validation task should succeed.")
        self.assertEqual(len(actions), 1)
        self.assertEqual(actions[0]['expected_class'], OBClasses.IdentityObject)
        self.assertEqual(actions[0]['node_path'], [first_node['id'], 'recipient'])

    def test_validate_linked_related_resource(self):
        first_node = {
            'id': 'http://example.com/1',
            'badge': 'http://example.com/2'
        }
        state = {'graph': [first_node]}

        task = add_task(
            VALIDATE_PROPERTY,
            node_id="http://example.com/1",
            prop_name="badge",
            required=True,
            prop_type=ValueTypes.ID,
            expected_class=OBClasses.BadgeClass,
            fetch=True
        )

        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "Property validation task should succeed.")
        self.assertEqual(len(actions), 1)
        self.assertEqual(actions[0]['expected_class'], OBClasses.BadgeClass)

    def test_many_validation_for_id_property(self):
        """
        When detect_and_validate_node_class (through _get_validation_actions)
        queue up actions, some configs may have many=True. This means single or multiple
        values should be accepted. If optional, empty lists should also be accepted.
        """
        first_node = {
            'id': '_:b0',
            'type': 'Assertion'
        }
        second_node = {
            'id': '_:b1',
            'narrative': 'Did cool stuff'
        }
        third_node = {
            'id': '_:b2',
            'narrative': 'Did more cool stuff'
        }
        state = {'graph': [first_node, second_node, third_node]}
        required = True

        task = add_task(
            VALIDATE_PROPERTY, node_id=first_node['id'], node_class=OBClasses.Assertion,
            prop_name='evidence', prop_type=ValueTypes.ID, required=required, many=True, fetch=False,
            expected_class=OBClasses.Evidence, allow_remote_url=True
        )
        result, message, actions = validate_property(state, task)
        self.assertFalse(result)
        self.assertTrue('Required property evidence not present' in message)

        first_node['evidence'] = 'http://youtube.com/avideoofmedoingthething'
        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "A single string URL value should be acceptable for evidence")

        first_node['evidence'] = 'notanurl'
        result, message, actions = validate_property(state, task)
        self.assertFalse(result, "A single string value that doesn't look like a URL is not acceptable as evidence")

        first_node['evidence'] = ['http://example.com/1', 'http://example.com/2']
        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "Multiple string URLs should be acceptable for evidence")

        first_node['evidence'] = ['_:b1', 'http://example.com/2']
        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "An embedded node and a URL should be acceptable evidence references")
        self.assertEqual(len(actions), 1)
        self.assertEqual(actions[0]['node_id'], '_:b1')

        # Test when many values are disallowed
        task = add_task(
            VALIDATE_PROPERTY, node_id=first_node['id'], node_class=OBClasses.Assertion,
            prop_name='evidence', prop_type=ValueTypes.ID, required=required, many=False, fetch=False,
            expected_class=OBClasses.Evidence, allow_remote_url=True
        )
        result, message, actions = validate_property(state, task)
        self.assertFalse(result, "Many values should be rejected when many=False")
        self.assertTrue('has more than the single allowed value' in message, "Error should mention many violation")

        task = add_task(
            VALIDATE_PROPERTY, node_id=first_node['id'], node_class=OBClasses.Assertion,
            prop_name='evidence', prop_type=ValueTypes.ID, required=required, fetch=False,
            expected_class=OBClasses.Evidence, allow_remote_url=True
        )
        result, message, actions = validate_property(state, task)
        self.assertFalse(result, "Many values should be rejected when many is not present")
        self.assertTrue('has more than the single allowed value' in message, "Error should mention many violation")

    def test_many_nested_validation_for_id_property(self):
        """
        When detect_and_validate_node_class (through _get_validation_actions)
        queue up actions, some configs may have many=True. This means single or multiple
        values should be accepted. If optional, empty lists should also be accepted.
        """
        first_node = {
            'id': '_:b0',
            'type': 'BadgeClass',
            'alignment': [
                {
                    'targetName': 'Alignment One',
                    'targetUrl': 'http://example.org/alignment1'
                }
            ]
        }
        state = {'graph': [first_node]}
        options = {'max_validation_depth': 3}
        required = True

        task = add_task(
            VALIDATE_PROPERTY, node_id=first_node['id'], node_class=OBClasses.BadgeClass,
            prop_name='alignment', prop_type=ValueTypes.ID, required=required, many=True, fetch=False,
            expected_class=OBClasses.AlignmentObject, allow_remote_url=False, depth=0
        )
        result, message, actions = validate_property(state, task)
        self.assertTrue(result, "Task the queues up individual class validators is successful")
        self.assertEqual(len(actions), 1)

        result, message, actions = task_named(actions[0]['name'])(state, actions[0], **options)
        self.assertTrue(result)
        for a in actions:
            self.assertEqual(a['node_path'], [first_node['id'], 'alignment', 0])
            next_result, next_message, next_actions = task_named(a['name'])(state, a)
            self.assertTrue(next_result)



class NodeTypeDetectionTasksTests(unittest.TestCase):
    def detect_assertion_type_from_node(self):
        node_data = json.loads(test_components['2_0_basic_assertion'])
        state = {'graph': [node_data]}
        task = add_task(DETECT_AND_VALIDATE_NODE_CLASS, node_id=node_data['id'])

        result, message, actions = detect_and_validate_node_class(state, task)
        self.assertTrue(result, "Type detection task should complete successfully.")
        self.assertEqual(len(actions), 5)

        issuedOn_task = [t for t in actions if t['prop_name'] == 'issuedOn'][0]
        self.assertEqual(issuedOn_task['prop_type'], ValueTypes.DATETIME)


class ClassValidationTaskTests(unittest.TestCase):
    def test_validate_identity_object_property_dependencies(self):
        first_node = {
            'id': 'http://example.com/1',
            'recipient': {
                'identity': 'sha256$c7ef86405ba71b85acd8e2e95166c4b111448089f2e1599f42fe1bba46e865c5',
                'type': 'email',
                'hashed': True,
                'salt': 'deadsea'
            }
        }
        state = {'graph': [first_node]}
        options = {'max_validation_depth': 3}

        task = add_task(
            VALIDATE_PROPERTY,
            node_id="http://example.com/1",
            prop_name="recipient",
            required=True,
            prop_type=ValueTypes.ID,
            expected_class=OBClasses.IdentityObject,
            depth=0
        )

        def run(cur_state, cur_task, expected_result, msg=''):
            result, message, actions = validate_property(cur_state, cur_task)
            self.assertTrue(result, "Property validation task should succeed.")
            self.assertEqual(len(actions), 1)
            self.assertEqual(actions[0]['expected_class'], OBClasses.IdentityObject)

            cur_task = actions[0]
            result, message, actions = task_named(cur_task['name'])(cur_state, cur_task, **options)
            self.assertTrue(result, "IdentityObject validation task discovery should succeed.")

            for cur_task in [a for a in actions if a.get('type') == ADD_TASK]:
                val_result, val_message, val_actions = task_named(cur_task['name'])(cur_state, cur_task)
                if not cur_task['name'] == IDENTITY_OBJECT_PROPERTY_DEPENDENCIES:
                    self.assertTrue(val_result, "Test {} should pass".format(cur_task['name']))
                else:
                    self.assertEqual(val_result, expected_result,
                                     "{} should be {}: {}".format(cur_task['name'], expected_result, msg))

        # Hashed and salted
        run(state, task, True, "Good working hashed identity")

        # Identity shouldn't look hashed if it says it isn't!
        first_node['recipient']['hashed'] = False
        run(state, task, False, "Identity looks hashed and smells fishy")

        # Hash doesn't match known types.
        first_node['recipient']['hashed'] = True
        first_node['recipient']['identity'] = "sha1billiion$abc123"
        run(state, task, False, "Hash doesn't match known types")
        first_node['recipient']['identity'] = "plaintextjane@example.com"
        run(state, task, False, "Identity shouldn't look like email if hashed is true")

    def test_criteria_property_dependency_validation(self):
        state = {
            'graph': [
                {'id': '_:b0'},
                {'id': '_:b1', 'narrative': 'Do cool stuff'},
                {'id': 'http://example.com/a', 'narrative': 'Do cool stuff'},
                {'id': 'http://example.com/b', 'name': 'Another property outside of Criteria class scope'},
            ]
        }
        task = add_task(CRITERIA_PROPERTY_DEPENDENCIES, node_id="_:b1")
        result, message, actions = criteria_property_dependencies(state, task)
        self.assertTrue(result)

        task = add_task(CRITERIA_PROPERTY_DEPENDENCIES, node_id="_:b1")
        result, message, actions = criteria_property_dependencies(state, task)
        self.assertTrue(result, "Evidence with blank node ID and narrative passes.")

        task = add_task(CRITERIA_PROPERTY_DEPENDENCIES, node_id="_:b0")
        result, message, actions = criteria_property_dependencies(state, task)
        self.assertFalse(result, "Evidence with just a blank node id fails.")

        task = add_task(CRITERIA_PROPERTY_DEPENDENCIES, node_id="_:b1")
        result, message, actions = criteria_property_dependencies(state, task)
        self.assertTrue(result, "Evidence with blank node ID and narrative passes.")

        task = add_task(CRITERIA_PROPERTY_DEPENDENCIES, node_id="http://example.com/b")
        result, message, actions = criteria_property_dependencies(state, task)
        self.assertTrue(result, "External URL with unknown properties passes.")

        task = add_task(CRITERIA_PROPERTY_DEPENDENCIES, node_id="http://example.com/a")
        result, message, actions = criteria_property_dependencies(state, task)
        self.assertTrue(result, "External URL and narrative passes")

    def test_run_criteria_task_discovery_and_validation(self):
        badgeclass_node = {'id': 'http://example.com/badgeclass', 'type': 'BadgeClass'}
        state = {
            'graph': [
                badgeclass_node,
                {'id': '_:b0'},
                {'id': '_:b1', 'narrative': 'Do cool stuff'},
                {'id': 'http://example.com/a', 'narrative': 'Do cool stuff'},
                {'id': 'http://example.com/b', 'name': 'Another property outside of Criteria class scope'},
            ]
        }
        options = {'max_validation_depth': 3}
        actions = [add_task(
            VALIDATE_PROPERTY,
            node_id=badgeclass_node['id'],
            prop_name="criteria",
            required=False,
            prop_type=ValueTypes.ID,
            expected_class=OBClasses.Criteria,
            depth=0
        )]
        badgeclass_node['criteria'] = 'http://example.com/a'

        for task in actions:
            if not task.get('type') == 'ADD_TASK':
                continue
            result, message, new_actions = task_named(task['name'])(state, task, **options)
            if new_actions:
                actions.extend(new_actions)
            self.assertTrue(result)

        self.assertTrue(CRITERIA_PROPERTY_DEPENDENCIES in [a.get('name') for a in actions])

    def test_run_criteria_task_discovery_and_validation_embedded(self):
        badgeclass_node = {'id': 'http://example.com/badgeclass', 'type': 'BadgeClass'}
        state = {
            'graph': [badgeclass_node]
        }
        options = {'max_validation_depth': 3}
        actions = [add_task(
            VALIDATE_PROPERTY,
            node_id=badgeclass_node['id'],
            prop_name="criteria",
            required=False,
            prop_type=ValueTypes.ID,
            expected_class=OBClasses.Criteria,
            depth=0
        )]
        badgeclass_node['criteria'] = {
            'narrative': 'Do the important things.'
        }

        for task in actions:
            if not task.get('type') == 'ADD_TASK':
                continue
            result, message, new_actions = task_named(task['name'])(state, task, **options)
            if new_actions:
                actions.extend(new_actions)
            self.assertTrue(result)

        self.assertTrue(CRITERIA_PROPERTY_DEPENDENCIES in [a.get('name') for a in actions])

    def test_many_criteria_disallowed(self):
        badgeclass_node = {'id': 'http://example.com/badgeclass', 'type': 'BadgeClass'}
        state = {'graph': [badgeclass_node]}
        options = {'max_validation_depth': 3}
        actions = [add_task(
            VALIDATE_PROPERTY,
            node_id=badgeclass_node['id'],
            prop_name="criteria",
            required=False,
            prop_type=ValueTypes.ID,
            expected_class=OBClasses.Criteria
        )]
        badgeclass_node['criteria'] = ['http://example.com/a', 'http://example.com/b']

        task = actions[0]
        result, message, new_actions = task_named(task['name'])(state, task, **options)
        self.assertFalse(result, "Validation should reject multiple criteria entries")
        self.assertTrue('has more than the single allowed value' in message)

    def _setUpEvidenceState(self):
        self.first_node = {
            'id': '_:b0',
            'type': 'Assertion'
        }
        self.second_node = {
            'id': '_:b1',
            'narrative': 'Did cool stuff'
        }
        self.third_node = {
            'id': '_:b2',
            'narrative': 'Earner did more cool stuff, like photography',
            'name': 'My porfolio item',
            'descrpition': 'A photo of earner\'s cat',
            'audience': 'Ages 0-99',
            'genre': 'Photography'
        }
        self.fourth_node = {
            'id': 'http://example.com/myblog/1',
            'narrative': 'Did more cool stuff'
        }
        self.empty_narrative = {
            'id': '_:b3'
        }
        self.state = {'graph': [
            self.first_node, self.second_node,
            self.third_node, self.empty_narrative,
            self.fourth_node
        ]}

    def _run(self, task_meta, expected_result, msg='', test_task='UNKNOWN'):
        options = {'max_validation_depth': 3}
        result, message, actions = validate_property(self.state, task_meta)
        self.assertTrue(result, "Property validation task should succeed.")
        self.assertEqual(len(actions), 1)

        task_meta = actions[0]
        result, message, actions = task_named(task_meta['name'])(self.state, task_meta, **options)
        self.assertTrue(result, "Class validation task discovery should succeed.")

        for task_meta in [a for a in actions if a.get('type') == ADD_TASK]:
            val_result, val_message, val_actions = task_named(task_meta['name'])(self.state, task_meta, **options)
            if not task_meta['name'] == test_task:
                self.assertTrue(val_result, "Test {} should pass".format(task_meta['name']))
            elif task_meta['name'] == test_task:
                self.assertEqual(
                    val_result, expected_result,
                    "{} should be {}: {}".format(task_meta['name'], expected_result, msg)
                )

    def test_evidence_class_validation(self):
        self._setUpEvidenceState()
        self.first_node['evidence'] = '_:b1'

        task = add_task(
            VALIDATE_PROPERTY,
            node_id="_:b0",
            prop_name="evidence",
            required=False,
            prop_type=ValueTypes.ID,
            expected_class=OBClasses.Evidence,
            depth=0
        )

        self._run(task, True, 'Single embedded complete evidence node passes')
        self.first_node['evidence'] = ['_:b3']
        self._run(task, True, 'Even an evidence list containing just a blank node should not fail.')
        self.first_node['evidence'] = 'http://example.com/myblog/1'
        self._run(task, True, 'Checks run even when the Evidence node ID is an external URL')

    def test_image_object_validation(self):
        image_data_node = {'id': 'data:image/gif;base64,R0lGODlhyAAiALM...DfD0QAADs=',
                           'type': 'schema:ImageObject'}
        image_uri_node = {'id': 'http://example.com/images/Badge',
                          'type': 'schema:ImageObject',
                          'caption': 'This is an image',
                          'author': 'http://example.com/users/ProfessorDale'}

        badgeclass_node = {'id': 'http://example.com/badgeclass',
                           'type': 'BadgeClass',
                           'image': image_data_node['id']}
        assertion_node = {'id': '_:b0',
                          'type': 'Assertion',
                          'image': image_uri_node['id']}

        badgeclass_state = {'graph': [badgeclass_node, image_data_node]}
        assertion_state = {'graph': [assertion_node, image_uri_node]}

        task = add_task(
            VALIDATE_PROPERTY,
            node_id=badgeclass_node['id'],
            prop_name='image',
            prop_required=False,
            prop_type=ValueTypes.DATA_URI_OR_URL,
            expected_class=OBClasses.Image
        )

        result, message, actions = validate_property(badgeclass_state, task)
        self.assertTrue(result, "Class validation of image property as data node in BadgeClass should succeed.")

        task['node_id']=assertion_node['id']
        result, message, actions = validate_property(assertion_state, task)
        self.assertTrue(result, "Class validation of image property as uri node in Assertion should succeed.")

    def test_alignment_object_validation(self):
        self.first_node = {'id': 'http://example.com/badge1'}
        self.state = {
            'graph': [
                self.first_node,
                {'id': '_:b0'},
                {'id': '_:b1', 'targetUrl': 'http://example.com/skill1', 'targetName': 'Cool Skill'},
            ]
        }
        self.first_node['alignment'] = '_:b1'

        task = add_task(
            VALIDATE_PROPERTY,
            node_id="http://example.com/badge1",
            prop_name="alignment",
            required=False,
            prop_type=ValueTypes.ID,
            expected_class=OBClasses.AlignmentObject,
            depth=0
        )

        self._run(task, True, 'Single embedded complete alignment node passes', test_task=None)

    @responses.activate
    def test_basic_badgeclass_validation(self):
        first_node = {
            '@context': OPENBADGES_CONTEXT_V2_DICT,
            'id': 'http://example.com/badgeclass',
            'type': 'BadgeClass',
            'name': 'Test Badge',
            'description': 'A badge of learning',
            'image': 'http://example.com/badgeimage',
            'criteria': '_:b0',
            'issuer': 'http://example.com/issuer',
            'tags': ['important', 'learning']
        }
        set_up_image_mock(first_node['image'])
        second_node = {'id': '_:b0', 'narrative': 'Do the important learning.'}
        state = {'graph': [first_node, second_node]}

        actions = _get_validation_actions(OBClasses.BadgeClass, 0, first_node['id'])
        results = []
        for action in actions:
            if action['type'] == 'ADD_TASK':
                results.append(
                    task_named(action['name'])(state, action)
                )
        print("RESULTS")
        print(results)
        self.assertTrue(all(i[0] for i in results))


class RdfTypeValidationTests(unittest.TestCase):
    @responses.activate
    def test_validate_in_context_string_type(self):
        set_up_context_mock()
        input_value = {
            '@context': OPENBADGES_CONTEXT_V2_DICT,
            'id': 'http://example.com/badge1',
            'type': 'BadgeClass',
            'name': 'Chumley'
        }

        test_types = (
            ('BadgeClass',                   True),
            (['Issuer', 'Extension'],        True),
            ('AlignmentObject',              True),
            ('http://example.com/CoolClass', True),
            ('NotAKnownClass',               False),
            ([],                             False),
            (['Issuer', 'UNKNOWN'],          False),
        )

        for type_value, expected_result in test_types:
            input_value['type'] = type_value
            compact_value = jsonld.compact(input_value, OPENBADGES_CONTEXT_V2_DICT)
            state = {'graph': [compact_value]}

            task_meta = add_task(
                VALIDATE_PROPERTY,
                node_id="http://example.com/badge1",
                prop_name="type",
                required=True,
                prop_type=ValueTypes.RDF_TYPE,
                many=True
            )
            result, message, actions = task_named(task_meta['name'])(state, task_meta)
            self.assertEqual(
                result, expected_result,
                "{} didn't meet expectation of result {}".format(type_value, expected_result))

    def test_rdf_property_default_applied(self):
        """
        Ensure that when a default is provided it is added to the class if there is no type declared.
        This occurs by a UPDATE_NODE task.
        """
        first_node = {
            "@context": OPENBADGES_CONTEXT_V2_DICT,
            'id': '_:b0',
            'narrative': 'Some criteria'
        }
        task_meta = add_task(
            VALIDATE_RDF_TYPE_PROPERTY,
            node_id='_:b0',
            prop_name='type',
            required=False,
            default='Criteria',
            prop_type=ValueTypes.RDF_TYPE,
            many=True
        )
        state = {'graph': [first_node]}
        result, message, actions = task_named(task_meta['name'])(state, task_meta)

        self.assertTrue(result)
        self.assertEqual(len(actions), 1)
        self.assertEqual(actions[0]['node_id'], '_:b0')
        self.assertEqual(actions[0]['type'], PATCH_NODE)

    def test_type_must_contain_one(self):
        """
        Ensure that a type property doesn't validate if it is required to contain at least one of
        one or more class options provided in task_meta['must_contain_one']
        """
        first_node = {
            "@context": OPENBADGES_CONTEXT_V2_DICT,
            'id': '_:b0',
            'name': 'Some Issuer'
        }
        task_meta = add_task(
            VALIDATE_RDF_TYPE_PROPERTY,
            node_id='_:b0',
            prop_name='type',
            required=True,
            prop_type=ValueTypes.RDF_TYPE,
            must_contain_one=['Issuer', 'Profile'],
            many=True
        )
        state = {'graph': [first_node]}
        result, message, actions = task_named(task_meta['name'])(state, task_meta)

        self.assertFalse(result)
        self.assertEqual(len(actions), 0)

        first_node['type'] = 'schema:Buffalo'
        result, message, actions = task_named(task_meta['name'])(state, task_meta)
        self.assertFalse(result, "Buffalo is not among allowable type values for this task.")
        self.assertTrue('does not have type among allowed values' in message)

        first_node['type'] = 'Issuer'
        result, message, actions = task_named(task_meta['name'])(state, task_meta)
        self.assertTrue(result, 'Issuer is one of the acceptable values for this task.')

        first_node['type'] = ['Issuer', 'schema:Buffalo']
        result, message, actions = task_named(task_meta['name'])(state, task_meta)
        self.assertTrue(result, 'Issuer as one of several values should be acceptable.')

        task_meta['must_contain_one'] = 'Assertion'
        first_node['type'] = 'A'
        result, message, actions = task_named(task_meta['name'])(state, task_meta)
        self.assertFalse(result, 'A portion of the acceptable value is not acceptable.')
        first_node['type'] = 'Assertion'
        result, message, actions = task_named(task_meta['name'])(state, task_meta)
        self.assertTrue(result, 'The exact string match between must_contain_one and type passes.')


class VerificationObjectValiationTests(unittest.TestCase):
    def test_bad_allowed_origins(self):
        issuer = {
            'id': 'http://example.org/issuer',
            'verification': {
                'allowedOrigins': ['example.com']
            }
        }
        state = {'graph':[issuer]}
        task = add_task(
            VALIDATE_PROPERTY, node_path=[issuer['id'], 'verification'],
            prop_name='allowedOrigins', prop_type=ValueTypes.URL_AUTHORITY, many=True
        )

        result, message, actions = run_task(state, task)
        self.assertTrue(result)
        issuer['verification']['allowedOrigins'] = ['http://mary.had/a-little-lamb']
        result, message, actions = run_task(state, task)
        self.assertFalse(result)


    def test_hosted_verification_object_in_assertion(self):
        assertion = {
            'type': 'Assertion',
            'id': 'http://example.com/assertion',
            'verification': '_:b0',
            'badge': 'http://example.com/badgeclass'
        }
        verification = {
            'id': '_:b0',
            'type': 'HostedBadge'
        }
        badge = {
            'id': 'http://example.com/badgeclass',
            'issuer': 'http://example.com/issuer'
        }
        issuer = {
            'id': 'http://example.com/issuer',
            'verification': '_:b1'
        }
        issuer_verification = {
            'id': '_:b1',
            'allowedOrigins': 'example.com'
        }
        state = {'graph': [assertion, verification, badge, issuer, issuer_verification]}
        task_meta = add_task(HOSTED_ID_IN_VERIFICATION_SCOPE, node_id=assertion['id'])

        result, message, actions = hosted_id_in_verification_scope(state, task_meta)
        self.assertTrue(result)

        # Use default value for allowedOrigins:
        del issuer_verification['allowedOrigins']
        result, message, actions = hosted_id_in_verification_scope(state, task_meta)
        self.assertTrue(result)

        # Ensure startsWith is used if present
        issuer_verification['startsWith'] = 'http://example.com/'
        result, message, actions = hosted_id_in_verification_scope(state, task_meta)
        self.assertTrue(result)
        issuer_verification['startsWith'] = 'http://example.com/NOT'
        result, message, actions = hosted_id_in_verification_scope(state, task_meta)
        self.assertFalse(result)

        # Handle multiple declared values for allowedOrigins
        issuer_verification['startsWith'] = ['http://example.com/NOT', 'http://example.com/ALSONOT']
        result, message, actions = hosted_id_in_verification_scope(state, task_meta)
        self.assertFalse(result)
        issuer_verification['startsWith'] = ['http://example.com/NOT', 'http://example.com/assert']
        result, message, actions = hosted_id_in_verification_scope(state, task_meta)
        self.assertTrue(result)

        # Use default policy
        del issuer['verification']
        result, message, actions = hosted_id_in_verification_scope(state, task_meta)
        self.assertTrue(result)


class AssertionTimeStampValidationTests(unittest.TestCase):
    def test_assertion_not_expired(self):
        machine_time_now = datetime.now(utc)
        an_hour_ago = machine_time_now - timedelta(hours=1)
        two_hours_ago = an_hour_ago - timedelta(hours=1)

        assertion = {
            'id': 'http://example.com/assertion',
            'issuedOn': two_hours_ago.isoformat()
        }
        state = {'graph': [assertion]}
        task_meta = add_task(ASSERTION_TIMESTAMP_CHECKS, node_id=assertion['id'])

        result, message, actions = assertion_timestamp_checks(state, task_meta)
        self.assertTrue(result)

        assertion['expires'] = an_hour_ago.isoformat()

        result, message, actions = assertion_timestamp_checks(state, task_meta)
        self.assertFalse(result)


    def test_assertion_not_expires_before_issue(self):
        machine_time_now = datetime.now(utc)
        an_hour_ago = machine_time_now - timedelta(hours=1)
        two_hours_ago = an_hour_ago - timedelta(hours=1)

        assertion = {
            'id': 'http://example.com/assertion',
            'issuedOn': an_hour_ago.isoformat(),
            'expires': two_hours_ago.isoformat()
        }
        state = {'graph': [assertion]}
        task_meta = add_task(ASSERTION_TIMESTAMP_CHECKS, node_id=assertion['id'])

        result, message, actions = assertion_timestamp_checks(state, task_meta)
        self.assertFalse(result, "Assertion that expires before issue should not be accepted.")
        self.assertTrue('expiration is prior to issue date' in message)

    def test_assertion_not_issued_in_future(self):
        machine_time_now = datetime.now(utc)
        an_hour_future = machine_time_now + timedelta(hours=1)

        assertion = {
            'id': 'http://example.com/assertion',
            'issuedOn': an_hour_future.isoformat()
        }
        state = {'graph': [assertion]}
        task_meta = add_task(ASSERTION_TIMESTAMP_CHECKS, node_id=assertion['id'])

        result, message, actions = assertion_timestamp_checks(state, task_meta)
        self.assertFalse(result, "Assertion issued in the future should not be accepted.")
        self.assertTrue('future in ')


class BadgeClassInputValidationTests(unittest.TestCase):
    @responses.activate
    def test_can_input_badgeclass(self):
        badgeclass = {
            '@context': OPENBADGES_CONTEXT_V2_DICT,
            'id': 'http://example.com/badgeclass1',
            'type': 'BadgeClass',
            'name': 'Example Badge',
            'description': 'An example',
            'criteria': 'http://example.com/criteria',
            'issuer': 'http://example.com/issuer1',
            'image': 'http://example.com/image1',
        }
        issuer = {
            '@context': OPENBADGES_CONTEXT_V2_DICT,
            'id': 'http://example.com/issuer1',
            'type': 'Issuer',
            'name': 'Example Issuer',
            'email': 'me@example.com',
            'url': 'http://example.com'
        }
        set_up_image_mock(badgeclass['image'])

        responses.add(responses.GET, badgeclass['id'], json=badgeclass)
        responses.add(responses.GET, issuer['id'], json=issuer)
        set_up_context_mock()

        results = verify('http://example.com/badgeclass1')
        self.assertTrue(results['report']['valid'])


class IssuerClassValidationTests(unittest.TestCase):
    def test_both_issuer_and_profile_queue_class_validation(self):
        issuer = {
            '@context': OPENBADGES_CONTEXT_V2_DICT,
            'id': 'http://example.com/issuer1',
            'type': 'Issuer',
            'url': 'http://example.com'
        }

        state = {'graph': [issuer]}
        options = {'max_validation_depth': 3}

        task_meta = add_task(VALIDATE_EXPECTED_NODE_CLASS, node_id=issuer['id'],
                             expected_class=issuer['type'], depth=0)

        result, message, actions = task_named(task_meta['name'])(state, task_meta, **options)
        self.assertTrue(result)

        task_meta = add_task(DETECT_AND_VALIDATE_NODE_CLASS, node_id=issuer['id'], depth=0)
        result, message, actions = task_named(task_meta['name'])(state, task_meta, **options)
        self.assertTrue(result)

    def test_issuer_warn_on_non_https_id(self):
        issuer = {
            '@context': OPENBADGES_CONTEXT_V2_DICT,
            'id': 'urn:uuid:2d391246-6e0d-4dab-906c-b29770bd7aa6',
            'type': 'Issuer',
            'url': 'http://example.com'
        }
        state = {'graph': [issuer]}
        options = {'max_validation_depth': 3}
        task_meta = add_task(ISSUER_PROPERTY_DEPENDENCIES, node_id=issuer['id'],
                             messageLevel=MESSAGE_LEVEL_WARNING)

        result, message, actions = task_named(task_meta['name'])(state, task_meta, **options)
        self.assertFalse(result)
        self.assertIn('HTTP', message)

        issuer['id'] = 'http://example.org/issuer1'
        task_meta['node_id'] = issuer['id']
        result, message, actions = task_named(task_meta['name'])(state, task_meta)
        self.assertTrue(result)

class ValidationDepthTests(unittest.TestCase):
    assertion = {
        '@context': OPENBADGES_CONTEXT_V2_URI,
        'type': 'Assertion',
        'id': 'http://example.com/assertion',
        'badge': 'http://example.com/badgeclass',
        'recipient': {
            'id': '_:b0',
            'identity': 'two@example.com',
            'type': 'email',
            'hashed': False
        },
        "verification": {
            "type": "HostedBadge"
        },
        'issuedOn': (datetime.now(utc) - timedelta(hours=1)).isoformat()
    }
    badgeclass = {
        '@context': OPENBADGES_CONTEXT_V2_URI,
        'id': 'http://example.com/badgeclass',
        'type': 'BadgeClass',
        'name': 'Example Badge',
        'description': 'An example',
        'criteria': 'http://example.com/criteria',
        'issuer': 'http://example.com/issuer',
        'image': 'http://example.com/image1',
    }
    endorsement = {
        '@context': OPENBADGES_CONTEXT_V2_URI,
        'id': 'http://example2.com/endorsement',
        'type': 'Endorsement',
        'claim': {
            'id': 'http://example.com/issuer',
            'endorsementComment': 'Pretty good issuer'
        },
        'issuedOn': '2017-10-01T00:00Z',
        'issuer': 'http://example2.com/issuer',
        'verification': {
            'type': "hosted"
        }
    }
    issuer = {
        '@context': OPENBADGES_CONTEXT_V2_URI,
        'id': 'http://example.com/issuer',
        'type': 'Issuer',
        'name': 'Example Issuer',
        'email': 'me@example.com',
        'url': 'http://example.com',
        'endorsement': endorsement['id']
    }
    endorsement_issuer_endorsement = {
        '@context': OPENBADGES_CONTEXT_V2_URI,
        'id': 'http://example3.com/endorsement',
        'type': 'Endorsement',
        'claim': {
            'id': 'http://example2.com/issuer',
            'endorsementComment': 'Pretty good endorsement issuer'
        },
        'issuedOn': '2017-10-01T00:00Z',
        'issuer': 'http://example3.com/issuer',
        'verification': {
            'type': "hosted"
        }
    }
    endorsement_issuer = {
        '@context': OPENBADGES_CONTEXT_V2_URI,
        'id': 'http://example2.com/issuer',
        'type': 'Issuer',
        'name': 'Example Issuer 2',
        'email': 'me@example2.com',
        'url': 'http://example2.com',
        'endorsement': endorsement_issuer_endorsement['id']
    }
    endorsement_issuer_endorsement_issuer = {
        '@context': OPENBADGES_CONTEXT_V2_URI,
        'id': 'http://example3.com/issuer',
        'type': 'Issuer',
        'name': 'Example Issuer 3',
        'email': 'me@example3.com',
        'url': 'http://example3.com'
    }

    @responses.activate
    def test_validation_depth(self):
        set_up_context_mock()

        for resource in [self.assertion, self.badgeclass, self.issuer, self.endorsement, self.endorsement_issuer]:
            responses.add(responses.GET, resource['id'], json=resource)
        set_up_image_mock(self.badgeclass['image'])

        results = verify(self.assertion['id'])
        self.assertTrue(results['report']['valid'])
        self.assertEqual(5, len(results['graph']), "The graph does not contain endorsement of issuer of endorsement.")

    @responses.activate
    def test_validation_depth_5(self):
        options = {'max_validation_depth': 5}
        set_up_context_mock()

        for resource in [self.assertion, self.badgeclass, self.issuer, self.endorsement, self.endorsement_issuer, self.endorsement_issuer_endorsement, self.endorsement_issuer_endorsement_issuer]:
            responses.add(responses.GET, resource['id'], json=resource)
        set_up_image_mock(self.badgeclass['image'])

        results = verify(self.assertion['id'], **options)
        self.assertTrue(results['report']['valid'])
        self.assertEqual(7, len(results['graph']), "The graph now contains all objects.")



from flask import Flask, redirect, render_template, request
import json
import six

from openbadges.verifier import verify


app = Flask(__name__)
app.config['MAX_CONTENT_LENGTH'] = 4 * 1024 * 1024  # 4mb file upload limit


def request_wants_json():
    best = request.accept_mimetypes.best_match(['application/json', 'text/html'])
    return best == 'application/json' and request.accept_mimetypes[best] > request.accept_mimetypes['text/html']


@app.route("/")
def home():
    provided_url = request.args.get('url', "")
    return render_template('index.html', url=provided_url)


@app.route("/results", methods=['GET'])
def result_get_redirect():
    return redirect('/')


@app.route("/results", methods=['POST'])
def results():
    data = request.get_json()
    profile = None
    if not data and isinstance(request.form.get('data'), six.string_types) or request.files:
        user_input = request.form['data']
        if 'image' in request.files and len(request.files['image'].filename):
            user_input = request.files['image']

        try:
            profile = json.loads(request.form.get('profile'))
        except (TypeError, ValueError):
            profile = None
    elif data:
        user_input = data.get('data')
        try:
            profile = data['profile']
            if isinstance(profile, six.string_types):
                profile = json.loads(profile)
        except (TypeError, ValueError, KeyError):
            pass

    verification_results = verify(user_input, recipient_profile=profile)

    if request_wants_json():
        return json.dumps(verification_results, indent=4), 200, {'Content-Type': 'application/json'}
    return render_template(
        'results.html', is_valid=verification_results.get('report', {}).get('valid'),
        error_count=verification_results.get('report', {}).get('errorCount'),
        results=json.dumps(verification_results, indent=4))


if __name__ == "__main__":
    app.run()



from ..actions.action_types import (RUN_VALIDATION_REPORT, SET_OPENBADGES_VERSION, SET_VALIDATION_SUBJECT,
                                    SET_VERIFIED_PROFILE)


def verification_report_reducer(state, action):
    if state is None:
        state = {}

    action_type = action.get('type')
    if action_type == RUN_VALIDATION_REPORT:
        # TODO
        return state
    elif action_type == SET_OPENBADGES_VERSION:
        state = state.copy()
        state['openBadgesVersion'] = action.get('version')
    elif action_type == SET_VALIDATION_SUBJECT:
        state = state.copy()
        state['validationSubject'] = action.get('node_id')
    elif action_type == SET_VERIFIED_PROFILE:
        state = state.copy()
        state['recipientProfile'] = {
            action.get('identityType', 'email'): action.get('identityValue')
        }

    return state



"""
These files are present to generate test mocks only. They will be moved out of the package into the testfiles directory
in a future 1.x release.
"""


class ExampleExtension(object):
    name = 'Example Extension'
    context_url = 'https://w3id.org/openbadges/extensions/exampleExtension/context.json'
    rdf_type = 'extensions:ExampleExtension'

    context_json = {
      "@context": {
        "obi": "https://w3id.org/openbadges#",
        "extensions": 'https://w3id.org/openbadges/extensions#',
        "exampleProperty": "http://schema.org/text"
      },
      "obi:validation": [
        {
          "obi:validatesType": "extensions:ExampleExtension",
          "obi:validationSchema": "https://w3id.org/openbadges/extensions/exampleExtension/schema.json"
        }
      ]
    }

    validation_schema = {
        "https://w3id.org/openbadges/extensions/exampleExtension/schema.json": {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "1.1 Open Badge Example Extension",
            "description": "An extension that allows you to add a single string exampleProperty to an extension object to represent some of your favorite text.",
            "type": "object",
            "properties": {
                "exampleProperty": {
                    "type": "string"
                }
            },
            "required": ["exampleProperty"]
        }
    }


class ApplyLink(object):
    name = 'Apply Link'
    context_url = 'http://w3id.org/openbadges/extensions/applyLinkExtension/context.json'
    rdf_type = 'extensions:ApplyLink'

    context_json = {
        "@context": {
            "obi": "https://w3id.org/openbadges#",
            "extensions": "https://w3id.org/openbadges/extensions#",
            "url": "extensions:applyLink"
        },
        "obi:validation": [
            {
                "obi:validatesType": "extensions:ApplyLink",
                "obi:validationSchema": "https://w3id.org/openbadges/extensions/applyLinkExtension/schema.json"
            }
        ]
    }

    validation_schema = {
        'https://w3id.org/openbadges/extensions/applyLinkExtension/schema.json': {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "Apply Link",
            "description": "An extension that allows you to add a single url to a web page providing information on how earners may apply for a badge.",
            "type": "object",
            "properties": {
                "url": {
                    "type": "string",
                    "format": "uri"
                }
            },
            "required": ["url"]
        }
    }


class GeoLocation(object):
    name = 'Geo Location'
    context_url = 'https://w3id.org/openbadges/extensions/geoCoordinatesExtension/context.json'
    rdf_type = 'extensions:GeoCoordinates'

    context_json = {
        "@context": {
            "obi": "https://w3id.org/openbadges#",
            "schema": "http://schema.org/",
            "extensions": "https://w3id.org/openbadges/extensions#",
            "geo": "schema:geo",
            "name": "schema:name",
            "description": "schema:description",
            "latitude": "schema:latitude",
            "longitude": "schema:longitude"
        },
        "obi:validation": [
            {
                "obi:validatesType": "extensions:GeoCoordinates",
                "obi:validationSchema": "https://w3id.org/openbadges/extensions/geoCoordinatesExtension/schema.json"
            }
        ]
    }

    validation_schema = {
        'https://w3id.org/openbadges/extensions/geoCoordinatesExtension/schema.json': {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "GeoCoordinates Open Badges Extension",
            "description": "An extension allowing for the addition of the geographic coordinates associated with a badge object. For example, geolocation could represent where a Badge Class is available, where a badge was earned or the location of an issuer. The required description property allows implementers to be more specific about the reason location is included. Implements Schema.org's Place class",
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "description": {"type": "string"},
                "geo": {
                    "type": "object",
                    "properties": {
                        "latitude": {"type": "number"},
                        "longitude": {"type": "number"}
                    },
                    "required": ["latitude", "longitude"]
                }
            },
            "required": ["description", "geo"]
        }
    }


ALL_KNOWN_EXTENSIONS = {
    'extensions:ApplyLink': ApplyLink,
    'extensions:ExampleExtension': ExampleExtension,
    'extensions:GeoCoordinates': GeoLocation
}
